"""
Google OpenIdConnect:
    https://python-social-auth.readthedocs.io/en/latest/backends/google.html
"""
from .open_id_connect import OpenIdConnectAuth
from .google import GoogleOAuth2


class GoogleOpenIdConnect(GoogleOAuth2, OpenIdConnectAuth):
    name = 'google-openidconnect'
    OIDC_ENDPOINT = 'https://accounts.google.com'
    # differs from value in discovery document
    # http://openid.net/specs/openid-connect-core-1_0.html#rfc.section.15.6.2
    ID_TOKEN_ISSUER = 'accounts.google.com'

    def user_data(self, access_token, *args, **kwargs):
        """Return user data from Google API"""
        return self.get_json(
            'https://www.googleapis.com/plus/v1/people/me/openIdConnect',
            params={'access_token': access_token, 'alt': 'json'}
        )



"""
ArcGIS OAuth2 backend
"""
from .oauth import BaseOAuth2


class ArcGISOAuth2(BaseOAuth2):
    name = 'arcgis'
    ID_KEY = 'username'
    AUTHORIZATION_URL = 'https://www.arcgis.com/sharing/rest/oauth2/authorize'
    ACCESS_TOKEN_URL = 'https://www.arcgis.com/sharing/rest/oauth2/token'
    ACCESS_TOKEN_METHOD = 'POST'
    EXTRA_DATA = [
        ('expires_in', 'expires_in')
    ]

    def get_user_details(self, response):
        """Return user details from ArcGIS account"""
        return {'username': response.get('username'),
                'email': response.get('email'),
                'fullname': response.get('fullName'),
                'first_name': response.get('firstName'),
                'last_name': response.get('lastName')}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            'https://www.arcgis.com/sharing/rest/community/self',
            params={
                'token': access_token,
                'f': 'json'
            }
        )



"""
Strava OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/strava.html
"""
from .oauth import BaseOAuth2


class StravaOAuth(BaseOAuth2):
    name = 'strava'
    AUTHORIZATION_URL = 'https://www.strava.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://www.strava.com/oauth/token'
    ACCESS_TOKEN_METHOD = 'POST'
    # Strava doesn't check for parameters in redirect_uri and directly appends
    # the auth parameters to it, ending with an URL like:
    # http://example.com/complete/strava?redirect_state=xxx?code=xxx&state=xxx
    # Check issue #259 for details.
    REDIRECT_STATE = False
    REVOKE_TOKEN_URL = 'https://www.strava.com/oauth/deauthorize'

    def get_user_id(self, details, response):
        return response['athlete']['id']

    def get_user_details(self, response):
        """Return user details from Strava account"""
        # because there is no usernames on strava
        username = response['athlete']['id']
        email = response['athlete'].get('email', '')
        fullname, first_name, last_name = self.get_user_names(
            first_name=response['athlete'].get('firstname', ''),
            last_name=response['athlete'].get('lastname', ''),
        )
        return {'username': str(username),
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name,
                'email': email}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json('https://www.strava.com/api/v3/athlete',
                             params={'access_token': access_token})

    def revoke_token_params(self, token, uid):
        params = super(StravaOAuth, self).revoke_token_params(token, uid)
        params['access_token'] = token
        return params



"""
Dropbox OAuth1 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/dropbox.html
"""
import warnings

from .oauth import BaseOAuth1, BaseOAuth2


class DropboxOAuth(BaseOAuth1):
    """Dropbox OAuth authentication backend"""
    name = 'dropbox'
    ID_KEY = 'uid'
    AUTHORIZATION_URL = 'https://www.dropbox.com/1/oauth/authorize'
    REQUEST_TOKEN_URL = 'https://api.dropbox.com/1/oauth/request_token'
    REQUEST_TOKEN_METHOD = 'POST'
    ACCESS_TOKEN_URL = 'https://api.dropbox.com/1/oauth/access_token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_URI_PARAMETER_NAME = 'oauth_callback'
    EXTRA_DATA = [
        ('id', 'id'),
        ('expires', 'expires')
    ]

    def __init__(self, *args, **kwargs):
        warnings.warn(
            'Dropbox V1 api is deprecated and will be shute down 2017-06-28 '
            'https://blogs.dropbox.com/developers/2016/06/api-v1-deprecated/',
            DeprecationWarning,
            stacklevel=2
        )
        super(DropboxOAuth, self).__init__(*args, **kwargs)

    def get_user_details(self, response):
        """Return user details from Dropbox account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get('display_name')
        )
        return {'username': str(response.get('uid')),
                'email': response.get('email'),
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json('https://api.dropbox.com/1/account/info',
                             auth=self.oauth_auth(access_token))


class DropboxOAuth2(BaseOAuth2):
    name = 'dropbox-oauth2'
    ID_KEY = 'uid'
    AUTHORIZATION_URL = 'https://www.dropbox.com/1/oauth2/authorize'
    ACCESS_TOKEN_URL = 'https://api.dropbox.com/1/oauth2/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ('uid', 'username'),
    ]

    def __init__(self, *args, **kwargs):
        warnings.warn(
            'Dropbox V1 api is deprecated and will be shute down 2017-06-28 '
            'https://blogs.dropbox.com/developers/2016/06/api-v1-deprecated/',
            DeprecationWarning,
            stacklevel=2
        )
        super(DropboxOAuth2, self).__init__(*args, **kwargs)

    def get_user_details(self, response):
        """Return user details from Dropbox account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get('display_name')
        )
        return {'username': str(response.get('uid')),
                'email': response.get('email'),
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            'https://api.dropbox.com/1/account/info',
            headers={'Authorization': 'Bearer {0}'.format(access_token)}
        )


class DropboxOAuth2V2(BaseOAuth2):
    name = 'dropbox-oauth2'
    ID_KEY = 'uid'
    AUTHORIZATION_URL = 'https://www.dropbox.com/oauth2/authorize'
    ACCESS_TOKEN_URL = 'https://api.dropboxapi.com/oauth2/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False

    def get_user_details(self, response):
        """Return user details from Dropbox account"""
        name = response.get('name')
        return {'username': str(response.get('account_id')),
                'email': response.get('email'),
                'fullname': name.get('display_name'),
                'first_name': name.get('given_name'),
                'last_name': name.get('surname')}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            'https://api.dropboxapi.com/2/users/get_current_account',
            headers={'Authorization': 'Bearer {0}'.format(access_token)},
            method='POST'
        )



"""
Kakao OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/kakao.html
"""
from .oauth import BaseOAuth2


class KakaoOAuth2(BaseOAuth2):
    """Kakao OAuth authentication backend"""
    name = 'kakao'
    AUTHORIZATION_URL = 'https://kauth.kakao.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://kauth.kakao.com/oauth/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ('properties', 'properties'),
    ]

    def get_user_id(self, details, response):
        return response['id']

    def get_user_details(self, response):
        """Return user details from Kakao account"""
        kaccount_email = response.get('kaccount_email', '')
        properties = response.get('properties', '')
        nickname = properties.get('nickname') if properties else ''
        return {
            'username': nickname,
            'email': kaccount_email,
            'fullname': nickname,
            'first_name': nickname[1:] if nickname else '',
            'last_name': nickname[0] if nickname else '',
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            'https://kapi.kakao.com/v1/user/me',
            headers={
                'Authorization': 'Bearer {0}'.format(access_token),
                'Content_Type': 'application/x-www-form-urlencoded;charset=utf-8',
            },
            params={'access_token': access_token}
        )

    def auth_complete_params(self, state=None):
        return {
            'grant_type': 'authorization_code',
            'code': self.data.get('code', ''),
            'client_id': self.get_key_and_secret()[0],
        }




import json

from jose import jwt
from httpretty import HTTPretty

from .oauth import OAuth2Test

JWK_KEY = {
    'kty': 'RSA',
    'd': 'ZmswNokEvBcxW_Kvcy8mWUQOQCBdGbnM0xR7nhvGHC-Q24z3XAQWlMWbsmGc_R1o' \
         '_F3zK7DBlc3BokdRaO1KJirNmnHCw5TlnBlJrXiWpFBtVglUg98-4sRRO0VWnGXK' \
         'JPOkBQ6b_DYRO3b0o8CSpWowpiV6HB71cjXTqKPZf-aXU9WjCCAtxVjfIxgQFu5I' \
         '-G1Qah8mZeY8HK_y99L4f0siZcbUoaIcfeWBhxi14ODyuSAHt0sNEkhiIVBZE7QZ' \
         'm-SEP1ryT9VAaljbwHHPmg7NC26vtLZhvaBGbTTJnEH0ZubbN2PMzsfeNyoCIHy4' \
         '4QDSpQDCHfgcGOlHY_t5gQ',
    'e': 'AQAB',
    'use': 'sig',
    'kid': 'foobar',
    'alg': 'RS256',
    'n': 'pUfcJ8WFrVue98Ygzb6KEQXHBzi8HavCu8VENB2As943--bHPcQ-nScXnrRFAUg8' \
         'H5ZltuOcHWvsGw_AQifSLmOCSWJAPkdNb0w0QzY7Re8NrPjCsP58Tytp5LicF0Ao' \
         'Ag28UK3JioY9hXHGvdZsWR1Rp3I-Z3nRBP6HyO18pEgcZ91c9aAzsqu80An9X4DA' \
         'b1lExtZorvcd5yTBzZgr-MUeytVRni2lDNEpa6OFuopHXmg27Hn3oWAaQlbymd4g' \
         'ifc01oahcwl3ze2tMK6gJxa_TdCf1y99Yq6oilmVvZJ8kwWWnbPE-oDmOVPVnEyT' \
         'vYVCvN4rBT1DQ-x0F1mo2Q'
}

JWK_PUBLIC_KEY = {key: value for key, value in JWK_KEY.items() if key != 'd'}

DOMAIN = 'foobar.auth0.com'


class Auth0OAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.auth0.Auth0OAuth2'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer',
        'expires_in': 86400,
        'id_token': jwt.encode({
            'nickname': 'foobar',
            'email': 'foobar@auth0.com',
            'name': 'John Doe',
            'picture': 'http://example.com/image.png',
            'sub': '123456',
            'iss': 'https://{}/'.format(DOMAIN),
        }, JWK_KEY, algorithm='RS256')
    })
    expected_username = 'foobar'
    jwks_url = 'https://foobar.auth0.com/.well-known/jwks.json'

    def extra_settings(self):
        settings = super(Auth0OAuth2Test, self).extra_settings()
        settings['SOCIAL_AUTH_' + self.name + '_DOMAIN'] = DOMAIN
        return settings

    def auth_handlers(self, start_url):
        HTTPretty.register_uri(HTTPretty.GET,
                               self.jwks_url,
                               body=json.dumps({'keys': [JWK_PUBLIC_KEY]}),
                               content_type='application/json')
        return super(Auth0OAuth2Test, self).auth_handlers(start_url)

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



from .oauth import BaseOAuth2


class TAOBAOAuth(BaseOAuth2):
    """Taobao OAuth authentication mechanism"""
    name = 'taobao'
    ID_KEY = 'taobao_user_id'
    ACCESS_TOKEN_METHOD = 'POST'
    AUTHORIZATION_URL = 'https://oauth.taobao.com/authorize'
    ACCESS_TOKEN_URL = 'https://oauth.taobao.com/token'

    def user_data(self, access_token, *args, **kwargs):
        """Return user data provided"""
        try:
            return self.get_json('https://eco.taobao.com/router/rest', params={
                'method': 'taobao.user.get',
                'fomate': 'json',
                'v': '2.0',
                'access_token': access_token
            })
        except ValueError:
            return None

    def get_user_details(self, response):
        """Return user details from Taobao account"""
        return {'username': response.get('taobao_user_nick')}



import base64

from .oauth import BaseOAuth2


class PushbulletOAuth2(BaseOAuth2):
    """pushbullet OAuth authentication backend"""
    name = 'pushbullet'
    EXTRA_DATA = [('id', 'id')]
    ID_KEY = 'username'
    AUTHORIZATION_URL = 'https://www.pushbullet.com/authorize'
    REQUEST_TOKEN_URL = 'https://api.pushbullet.com/oauth2/token'
    ACCESS_TOKEN_URL = 'https://api.pushbullet.com/oauth2/token'
    ACCESS_TOKEN_METHOD = 'POST'
    STATE_PARAMETER = False

    def get_user_details(self, response):
        return {'username': response.get('access_token')}

    def get_user_id(self, details, response):
        auth = 'Basic {0}'.format(base64.b64encode(details['username']))
        return self.get_json('https://api.pushbullet.com/v2/users/me',
                             headers={'Authorization': auth})['iden']



"""
Twitch OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/twitch.html
"""
from .oauth import BaseOAuth2


class TwitchOAuth2(BaseOAuth2):
    """Twitch OAuth authentication backend"""
    name = 'twitch'
    ID_KEY = '_id'
    AUTHORIZATION_URL = 'https://api.twitch.tv/kraken/oauth2/authorize'
    ACCESS_TOKEN_URL = 'https://api.twitch.tv/kraken/oauth2/token'
    ACCESS_TOKEN_METHOD = 'POST'
    DEFAULT_SCOPE = ['user_read']
    REDIRECT_STATE = False

    def get_user_details(self, response):
        return {
            'username': response.get('name'),
            'email': response.get('email'),
            'first_name': '',
            'last_name': ''
        }

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json(
            'https://api.twitch.tv/kraken/user/',
            params={'oauth_token': access_token}
        )



"""
NGP VAN's `ActionID` Provider

http://developers.ngpvan.com/action-id
"""
from openid.extensions import ax

from .open_id import OpenIdAuth


class ActionIDOpenID(OpenIdAuth):
    """
    NGP VAN's ActionID OpenID 1.1 authentication backend
    """
    name = 'actionid-openid'
    URL = 'https://accounts.ngpvan.com/Home/Xrds'
    USERNAME_KEY = 'email'

    def get_ax_attributes(self):
        """
        Return the AX attributes that ActionID responds with, as well as the
        user data result that it must map to.
        """
        return [
            ('http://openid.net/schema/contact/internet/email', 'email'),
            ('http://openid.net/schema/contact/phone/business', 'phone'),
            ('http://openid.net/schema/namePerson/first', 'first_name'),
            ('http://openid.net/schema/namePerson/last', 'last_name'),
            ('http://openid.net/schema/namePerson', 'fullname'),
        ]

    def setup_request(self, params=None):
        """
        Setup the OpenID request

        Because ActionID does not advertise the availiability of AX attributes
        nor use standard attribute aliases, we need to setup the attributes
        manually instead of rely on the parent OpenIdAuth.setup_request()
        """
        request = self.openid_request(params)

        fetch_request = ax.FetchRequest()
        fetch_request.add(ax.AttrInfo(
            'http://openid.net/schema/contact/internet/email',
            alias='ngpvanemail',
            required=True
        ))

        fetch_request.add(ax.AttrInfo(
            'http://openid.net/schema/contact/phone/business',
            alias='ngpvanphone',
            required=False
        ))
        fetch_request.add(ax.AttrInfo(
            'http://openid.net/schema/namePerson/first',
            alias='ngpvanfirstname',
            required=False
        ))
        fetch_request.add(ax.AttrInfo(
            'http://openid.net/schema/namePerson/last',
            alias='ngpvanlastname',
            required=False
        ))
        request.addExtension(fetch_request)

        return request



"""
LinkedIn OAuth1 and OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/linkedin.html
"""
from .oauth import BaseOAuth1, BaseOAuth2


class BaseLinkedinAuth(object):
    EXTRA_DATA = [('id', 'id'),
                  ('expires_in', 'expires'),
                  ('first-name', 'first_name', True),
                  ('last-name', 'last_name', True),
                  ('firstName', 'first_name', True),
                  ('lastName', 'last_name', True)]
    USER_DETAILS = 'https://api.linkedin.com/v1/people/~:({0})'

    def get_user_details(self, response):
        """Return user details from Linkedin account"""
        fullname, first_name, last_name = self.get_user_names(
            first_name=response['firstName'],
            last_name=response['lastName']
        )
        email = response.get('emailAddress', '')
        return {'username': first_name + last_name,
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name,
                'email': email}

    def user_details_url(self):
        # use set() since LinkedIn fails when values are duplicated
        fields_selectors = list(set(['first-name', 'id', 'last-name'] +
                                self.setting('FIELD_SELECTORS', [])))
        # user sort to ease the tests URL mocking
        fields_selectors.sort()
        fields_selectors = ','.join(fields_selectors)
        return self.USER_DETAILS.format(fields_selectors)

    def user_data_headers(self):
        lang = self.setting('FORCE_PROFILE_LANGUAGE')
        if lang:
            return {
                'Accept-Language': lang if lang is not True
                                        else self.strategy.get_language()
            }


class LinkedinOAuth(BaseLinkedinAuth, BaseOAuth1):
    """Linkedin OAuth authentication backend"""
    name = 'linkedin'
    SCOPE_SEPARATOR = '+'
    AUTHORIZATION_URL = 'https://www.linkedin.com/uas/oauth/authenticate'
    REQUEST_TOKEN_URL = 'https://api.linkedin.com/uas/oauth/requestToken'
    ACCESS_TOKEN_URL = 'https://api.linkedin.com/uas/oauth/accessToken'

    def user_data(self, access_token, *args, **kwargs):
        """Return user data provided"""
        return self.get_json(
            self.user_details_url(),
            params={'format': 'json'},
            auth=self.oauth_auth(access_token),
            headers=self.user_data_headers()
        )

    def unauthorized_token(self):
        """Makes first request to oauth. Returns an unauthorized Token."""
        scope = self.get_scope() or ''
        if scope:
            scope = '?scope=' + self.SCOPE_SEPARATOR.join(scope)
        return self.request(self.REQUEST_TOKEN_URL + scope,
                            params=self.request_token_extra_arguments(),
                            auth=self.oauth_auth()).text


class LinkedinOAuth2(BaseLinkedinAuth, BaseOAuth2):
    name = 'linkedin-oauth2'
    SCOPE_SEPARATOR = ' '
    AUTHORIZATION_URL = 'https://www.linkedin.com/uas/oauth2/authorization'
    ACCESS_TOKEN_URL = 'https://www.linkedin.com/uas/oauth2/accessToken'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False

    def user_data(self, access_token, *args, **kwargs):
        headers = self.user_data_headers() or {}
        headers['Authorization'] = 'Bearer {access_token}'.format(
            access_token=access_token
        )
        return self.get_json(
            self.user_details_url(),
            params={'format': 'json'},
            headers=headers
        )

    def request_access_token(self, *args, **kwargs):
        # LinkedIn expects a POST request with querystring parameters, despite
        # the spec http://tools.ietf.org/html/rfc6749#section-4.1.3
        kwargs['params'] = kwargs.pop('data')
        return super(LinkedinOAuth2, self).request_access_token(
            *args, **kwargs
        )


class LinkedinMobileOAuth2(LinkedinOAuth2):
    name = 'linkedin-mobile-oauth2'

    def user_data(self, access_token, *args, **kwargs):
        headers = self.user_data_headers()
        if not headers:
            headers = {}
        headers['Authorization'] = 'Bearer ' + access_token
        headers['x-li-src'] = 'msdk'
        return self.get_json(
            self.user_details_url(),
            params={'format': 'json'},
            headers=headers
        )



import jwt

from social_core.backends.oauth import BaseOAuth2


class KeycloakOAuth2(BaseOAuth2):  # pylint: disable=abstract-method
    """Keycloak OAuth2 authentication backend

    This backend has been tested working with a standard Keycloak installation,
    but you might have to specialize it and tune the parameters per your configuration.

    This setup specializes the OAuth2 backend which, strictly speaking,
    offers authorization without authentication capabilities.

    Keycloak does offer a full OpenID Connect implementation,
    but the implementation is rather labor intensive to implement.

    This backend is configured to get an access token instead, and assume that the
    access token contains the necessary user details for authentication.

    The integrity of the authentication process is followed by public key verification
    for the `access_token` along with OpenID Connect specification `aud` field checking.

    To set up, please take the following steps:

    1. Create a new Keycloak client in the Clients section.

    2. Configure the following parameters in the Client setup:

        Settings > Client ID (copy to settings as `KEY` value)
        Credentials > Client Authenticator > Secret (copy to settings as `SECRET` value)

    3. For the tokens to work with the JWT setup the following configuration has to be made in Keycloak:

        Settings > Access Type > confidential
        Settings > Fine Grain OpenID Connect Configuration > User Info Signed Response Algorithm > RS256
        Settings > Fine Grain OpenID Connect Configuration > Request Object Signature Algorithm > RS256

    4. Get the public key (copy to settings as `PUBLIC_KEY` value) to be used with the backend:

        Realm Settings > Keys > Public key

    5. Configure access token fields are configured via the Keycloak Client mappers:

        Clients > Client ID > Mappers

    They have to include at least the `ID_KEY` value and the dictionary keys defined in the `get_user_details` method.

    6. Configure your web backend. Example setting values for Django settings could be:

        SOCIAL_AUTH_KEYCLOAK_KEY = 'example'
        SOCIAL_AUTH_KEYCLOAK_SECRET = '1234abcd-1234-abcd-1234-abcd1234adcd'
        SOCIAL_AUTH_KEYCLOAK_PUBLIC_KEY = 'pempublickeythatis2048bitsinbase64andhaseg392characters'
        SOCIAL_AUTH_KEYCLOAK_AUTHORIZATION_URL = 'https://sso.example.com/auth/realms/example/protocol/openid-connect/auth'
        SOCIAL_AUTH_KEYCLOAK_ACCESS_TOKEN_URL = 'https://sso.example.com/auth/realms/example/protocol/openid-connect/token'

    7. The default behaviour is to associate users via username field, but you can change the key with e.g.

            SOCIAL_AUTH_KEYCLOAK_ID_KEY = 'email'

    Please make sure your Keycloak user database and Django user database do not conflict
    and that there is no risk of user account hijacking by false account association.
    """

    name = 'keycloak'
    ID_KEY = 'username'
    ACCESS_TOKEN_METHOD = 'POST'

    def authorization_url(self):
        return self.setting('AUTHORIZATION_URL')

    def access_token_url(self):
        return self.setting('ACCESS_TOKEN_URL')

    def audience(self):
        return self.setting('KEY')

    def algorithm(self):
        return self.setting('ALGORITHM', default='RS256')

    def public_key(self):
        return '\n'.join([
            '-----BEGIN PUBLIC KEY-----',
            self.setting('PUBLIC_KEY'),
            '-----END PUBLIC KEY-----',
        ])

    def user_data(self, access_token, *args, **kwargs):  # pylint: disable=unused-argument
        """Decode user data from the access_token

        You can specialize this method to e.g. get information
        from the Keycloak backend if you do not want to include
        the user information in the access_token.
        """

        return jwt.decode(
            access_token,
            key=self.public_key(),
            algorithms=self.algorithm(),
            audience=self.audience(),
        )

    def get_user_details(self, response):
        """Map fields in user_data into Django User fields
        """

        return {
            'username': response.get('preferred_username'),
            'email': response.get('email'),
            'fullname': response.get('name'),
            'first_name': response.get('given_name'),
            'last_name': response.get('family_name'),
        }

    def get_user_id(self, details, response):
        """Get and associate Django User by the field indicated by ID_KEY
        """

        return details.get(self.ID_KEY)



import json

from httpretty import HTTPretty

from six.moves.urllib_parse import urlencode

from ...exceptions import AuthForbidden
from .oauth import OAuth1Test, OAuth2Test


class UberOAuth2Test(OAuth2Test):
    user_data_url = 'https://api.uber.com/v1/me'
    backend_path = 'social_core.backends.uber.UberOAuth2'
    expected_username = 'foo@bar.com'

    user_data_body = json.dumps({
        "first_name": "Foo",
        "last_name": "Bar",
        "email": "foo@bar.com",
        "picture": "https://",
        "promo_code": "barfoo",
        "uuid": "91d81273-45c2-4b57-8124-d0165f8240c0"
    })

    access_token_body = json.dumps({
        "access_token": "EE1IDxytP04tJ767GbjH7ED9PpGmYvL",
        "token_type": "Bearer",
        "expires_in": 2592000,
        "refresh_token": "Zx8fJ8qdSRRseIVlsGgtgQ4wnZBehr",
        "scope": "profile history request"
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Legacy Username backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/username.html
"""
from .legacy import LegacyAuth


class UsernameAuth(LegacyAuth):
    name = 'username'
    ID_KEY = 'username'
    EXTRA_DATA = ['username']



"""
Coursera OAuth2 backend, docs at:
    https://tech.coursera.org/app-platform/oauth2/
"""
from .oauth import BaseOAuth2


class CourseraOAuth2(BaseOAuth2):
    """Coursera OAuth2 authentication backend"""
    name = 'coursera'
    ID_KEY = 'username'
    AUTHORIZATION_URL = 'https://accounts.coursera.org/oauth2/v1/auth'
    ACCESS_TOKEN_URL = 'https://accounts.coursera.org/oauth2/v1/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False
    SCOPE_SEPARATOR = ','
    DEFAULT_SCOPE = ['view_profile']

    def _get_username_from_response(self, response):
        elements = response.get('elements', [])
        for element in elements:
            if 'id' in element:
                return element.get('id')

        return None

    def get_user_details(self, response):
        """Return user details from Coursera account"""
        return {'username': self._get_username_from_response(response)}

    def get_user_id(self, details, response):
        """Return a username prepared in get_user_details as uid"""
        return details.get(self.ID_KEY)

    def user_data(self, access_token, *args, **kwargs):
        """Load user data from the service"""
        return self.get_json(
            'https://api.coursera.org/api/externalBasicProfiles.v1?q=me',
            headers=self.get_auth_header(access_token)
        )

    def get_auth_header(self, access_token):
        return {'Authorization': 'Bearer {0}'.format(access_token)}



""" SciStarter OAuth2 Auth """

from six.moves.urllib_parse import unquote, urlencode

from .oauth import BaseOAuth2


class SciStarterOAuth2(BaseOAuth2):
    name = 'scistarter'
    ID_KEY = 'email'
    SCOPE_PARAMETER_NAME = 'scope'
    DEFAULT_SCOPE = ['login', 'extensive']
    SCOPE_SEPARATOR = ' '
    AUTHORIZATION_URL = 'https://scistarter.com/authorize'
    ACCESS_TOKEN_URL = 'https://scistarter.com/token?key={key}'
    ACCESS_TOKEN_METHOD = 'POST'
    USER_ACCESS_URL = \
        'https://scistarter.com/api/user_info?client_id={clientid}&key={key}'
    REFRESH_TOKEN_URL = None
    RESPONSE_TYPE = 'code'
    STATE_PARAMETER = True
    REDIRECT_STATE = True
    EXTRA_DATA = [
        ('profile_id', 'profile_id'),
        ('expires', 'expires')
    ]

    def get_redirect_uri(self, state=None):
        """Build redirect with redirect_state parameter."""
        return self.redirect_uri.rstrip('/')

    def authorization_url(self):
        return self.AUTHORIZATION_URL

    def get_user_details(self, response):
        return {
            'username': response.get('handle'),
            'email': response.get('email') or '',
            'first_name': response.get('first_name'),
            'last_name': response.get('last_name')
        }

    def user_data(self, access_token, *args, **kwards):
        client_id, client_secret = self.get_key_and_secret()
        return self.get_json(
            self.USER_ACCESS_URL.format(clientid=client_id, key=client_secret),
            headers={
                'Authorization': 'Bearer ' + access_token
            }
        )

    def access_token(self, token):
        """Return request for access token value"""
        return self.get_querystring(self.access_token_url(),
                                    auth=self.oauth_auth(token),
                                    method=self.ACCESS_TOKEN_METHOD)

    def access_token_url(self):
        client_id, client_secret = self.get_key_and_secret()
        return self.ACCESS_TOKEN_URL.format(key=client_secret)



import json

from .oauth import OAuth2Test


class AsanaOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.asana.AsanaOAuth2'
    user_data_url = 'https://app.asana.com/api/1.0/users/me'
    expected_username = 'erlich@bachmanity.com'
    access_token_body = json.dumps({
        'access_token': 'aviato',
        'token_type': 'bearer'
    })
    # https://asana.com/developers/api-reference/users
    user_data_body = json.dumps({
        'data': {
            'id': 12345,
            'name': 'Erlich Bachman',
            'email': 'erlich@bachmanity.com',
            'photo': None,
            'workspaces': [
                {
                    'id': 123456,
                    'name': 'Pied Piper'
                }
            ]
        }
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json
from .open_id_connect import OpenIdConnectAuth


class PixelPinOpenIDConnect(OpenIdConnectAuth):
    """PixelPin OpenID Connect authentication backend"""
    name = 'pixelpin-openidconnect'
    ID_KEY = 'sub'
    AUTHORIZATION_URL = 'https://login.pixelpin.io/connect/authorize'
    ACCESS_TOKEN_URL = 'https://login.pixelpin.io/connect/token'
    OIDC_ENDPOINT = 'https://login.pixelpin.io'
    JWKS_URI = 'https://login.pixelpin.io/.well-known/jwks'
    ACCESS_TOKEN_METHOD = 'POST'
    REQUIRES_EMAIL_VALIDATION = False

    def get_user_details(self, response):
        """Return user details from PixelPin account"""
        first_name = response.get('given_name')
        last_name = response.get('family_name')
        sub = response.get('sub')

        username = first_name + last_name + sub

        return {'username': username,
                'email': response.get('email'),
                'fullname': first_name + ' ' + last_name,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json(
            'https://login.pixelpin.io/connect/userinfo',
            headers={
                'Authorization': 'Bearer {0}'.format(access_token)
            }
        )



"""
ORCID OAuth2 Application backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/orcid.html
"""
from .oauth import BaseOAuth2


class ORCIDOAuth2(BaseOAuth2):
    """ORCID OAuth2 authentication backend"""
    name = 'orcid'
    ID_KEY = 'orcid'
    AUTHORIZATION_URL = 'https://orcid.org/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://orcid.org/oauth/token'
    USER_DATA_URL = 'https://pub.orcid.org/v2.0/{}'
    DEFAULT_SCOPE = ['/authenticate']
    ACCESS_TOKEN_METHOD = 'POST'
    EXTRA_DATA = [
        ('orcid', 'id'),
        ('expires_in', 'expires'),
        ('refresh_token', 'refresh_token')
    ]

    def auth_params(self, state=None):
        params = super(ORCIDOAuth2, self).auth_params(state)
        return params

    def get_user_details(self, response):
        """Return user details from ORCID account"""
        fullname= response.get('name', '')
        first_name = last_name = email = ''
        person = response.get('person')
        if person:
            name = person.get('name')
            if name:
                first_name = name.get('given-names', {}).get('value', '')
                last_name = name.get('family-name', {}).get('value', '')

            emails = person.get('emails')
            if emails:
                emails_list = emails.get('email')
                if emails_list and len(emails_list) > 0:
                    email = emails_list[0].get('email', '')

                    if len(emails_list) > 1:
                        for email_dict in emails_list:
                            if email_dict.get('primary','') == True:
                                email = email_dict.get('email', '')
                                break
                    else:
                        mail = emails_list[0].get('email', '')

        return {'username': response.get('orcid'),
                'email': email,
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        params = self.setting('PROFILE_EXTRA_PARAMS', {})
        params['access_token'] = access_token
        try:
            return self.get_json(self.USER_DATA_URL.format(
                            kwargs['response']['orcid']),
                            headers={'Content-Type': 'application/json'},
                            params=params)
        except ValueError as e:
            return None


class ORCIDOAuth2Sandbox(ORCIDOAuth2):
    """ORCID OAuth2 Sandbox authentication backend"""
    name = 'orcid-sandbox'
    AUTHORIZATION_URL = 'https://sandbox.orcid.org/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://sandbox.orcid.org/oauth/token'
    USER_DATA_URL = 'https://pub.sandbox.orcid.org/v2.0/{}'


class ORCIDMemberOAuth2(ORCIDOAuth2):
    """ORCID OAuth2 authentication backend that uses ORCID Member API"""
    USER_DATA_URL = 'https://api.orcid.org/v2.0/{}'
    DEFAULT_SCOPE = ['/authenticate', '/read-limited']


class ORCIDMemberOAuth2Sandbox(ORCIDOAuth2Sandbox):
    """ORCID OAuth2 Sandbox authentication backend that uses ORCID Member Sandbox API"""
    USER_DATA_URL = 'https://api.sandbox.orcid.org/v2.0/{}'
    DEFAULT_SCOPE = ['/authenticate', '/read-limited']



import json
from .oauth import OAuth2Test


class TwitchOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.twitch.TwitchOAuth2'
    user_data_url = 'https://api.twitch.tv/kraken/user/'
    expected_username = 'test_user1'
    access_token_body = json.dumps({
        'access_token': 'foobar',
    })
    user_data_body = json.dumps({
        'type': 'user',
        'name': 'test_user1',
        'created_at': '2011-06-03T17:49:19Z',
        'updated_at': '2012-06-18T17:19:57Z',
        '_links': {
            'self': 'https://api.twitch.tv/kraken/users/test_user1'
        },
        'logo': 'http://static-cdn.jtvnw.net/jtv_user_pictures/'
                'test_user1-profile_image-62e8318af864d6d7-300x300.jpeg',
        '_id': 22761313,
        'display_name': 'test_user1',
        'email': 'asdf@asdf.com',
        'partnered': True,
        'bio': 'test bio woo I\'m a test user'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import OAuth2Test


class RedditOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.reddit.RedditOAuth2'
    user_data_url = 'https://oauth.reddit.com/api/v1/me.json'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'name': 'foobar',
        'created': 1203420772.0,
        'access_token': 'foobar-token',
        'created_utc': 1203420772.0,
        'expires_in': 3600.0,
        'link_karma': 34,
        'token_type': 'bearer',
        'comment_karma': 167,
        'over_18': True,
        'is_gold': False,
        'is_mod': True,
        'scope': 'identity',
        'has_verified_email': False,
        'id': '33bma',
        'refresh_token': 'foobar-refresh-token'
    })
    user_data_body = json.dumps({
        'name': 'foobar',
        'created': 1203420772.0,
        'created_utc': 1203420772.0,
        'link_karma': 34,
        'comment_karma': 167,
        'over_18': True,
        'is_gold': False,
        'is_mod': True,
        'has_verified_email': False,
        'id': '33bma'
    })
    refresh_token_body = json.dumps({
        'access_token': 'foobar-new-token',
        'token_type': 'bearer',
        'expires_in': 3600.0,
        'refresh_token': 'foobar-new-refresh-token',
        'scope': 'identity'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()

    def refresh_token_arguments(self):
        uri = self.strategy.build_absolute_uri('/complete/reddit/')
        return {'redirect_uri': uri}

    def test_refresh_token(self):
        user, social = self.do_refresh_token()
        self.assertEqual(social.extra_data['access_token'], 'foobar-new-token')



# -*- coding: utf-8 -*-
"""
Pinterest OAuth2 backend, docs at:
    https://developers.pinterest.com/docs/api/authentication/
"""

from __future__ import unicode_literals

import ssl

from .oauth import BaseOAuth2


class PinterestOAuth2(BaseOAuth2):
    name = 'pinterest'
    ID_KEY = 'user_id'
    AUTHORIZATION_URL = 'https://api.pinterest.com/oauth/'
    ACCESS_TOKEN_URL = 'https://api.pinterest.com/v1/oauth/token'
    REDIRECT_STATE = False
    ACCESS_TOKEN_METHOD = 'POST'
    SSL_PROTOCOL = ssl.PROTOCOL_TLSv1

    def user_data(self, access_token, *args, **kwargs):
        response = self.get_json('https://api.pinterest.com/v1/me/',
                                 params={'access_token': access_token})

        if 'data' in response:
            username = response['data']['url'].strip('/').split('/')[-1]
            response = {
                'user_id': response['data']['id'],
                'first_name': response['data']['first_name'],
                'last_name': response['data']['last_name'],
                'username': username,
            }
        return response

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(
            first_name=response['first_name'],
            last_name=response['last_name'])

        return {'username': response.get('username'),
                'email': None,
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}



"""
Readability OAuth1 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/readability.html
"""
from .oauth import BaseOAuth1


READABILITY_API = 'https://www.readability.com/api/rest/v1'


class ReadabilityOAuth(BaseOAuth1):
    """Readability OAuth authentication backend"""
    name = 'readability'
    ID_KEY = 'username'
    AUTHORIZATION_URL = '{0}/oauth/authorize/'.format(READABILITY_API)
    REQUEST_TOKEN_URL = '{0}/oauth/request_token/'.format(READABILITY_API)
    ACCESS_TOKEN_URL = '{0}/oauth/access_token/'.format(READABILITY_API)
    EXTRA_DATA = [('date_joined', 'date_joined'),
                  ('kindle_email_address', 'kindle_email_address'),
                  ('avatar_url', 'avatar_url'),
                  ('email_into_address', 'email_into_address')]

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(
            first_name=response['first_name'],
            last_name=response['last_name']
        )
        return {'username': response['username'],
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token):
        return self.get_json(READABILITY_API + '/users/_current',
                             auth=self.oauth_auth(access_token))



from requests.auth import HTTPBasicAuth
from ..utils import handle_http_errors
from .oauth import BaseOAuth2


class JustGivingOAuth2(BaseOAuth2):
    """Just Giving OAuth authentication backend"""
    name = 'justgiving'
    ID_KEY = 'userId'
    AUTHORIZATION_URL = 'https://identity.justgiving.com/connect/authorize'
    ACCESS_TOKEN_URL = 'https://identity.justgiving.com/connect/token'
    ACCESS_TOKEN_METHOD = 'POST'
    USER_DATA_URL = 'https://api.justgiving.com/v1/account'
    DEFAULT_SCOPE = ['openid', 'account', 'profile', 'email', 'fundraise']

    def get_user_details(self, response):
        """Return user details from Just Giving account"""
        fullname, first_name, last_name = self.get_user_names(
            '',
            response.get('firstName'),
            response.get('lastName'))
        return {
            'username': response.get('email'),
            'email': response.get('email'),
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        key, secret = self.get_key_and_secret()
        return self.get_json(self.USER_DATA_URL, headers={
            'Authorization': 'Bearer {0}'.format(access_token),
            'Content-Type': 'application/json',
            'x-application-key': secret,
            'x-api-key': key
        })

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        state = self.validate_state()
        self.process_error(self.data)

        key, secret = self.get_key_and_secret()
        response = self.request_access_token(
            self.access_token_url(),
            data=self.auth_complete_params(state),
            headers=self.auth_headers(),
            auth=HTTPBasicAuth(key, secret),
            method=self.ACCESS_TOKEN_METHOD
        )
        self.process_error(response)
        return self.do_auth(response['access_token'], response=response,
                            *args, **kwargs)



from .oauth import BaseOAuth2


class GoClioOAuth2(BaseOAuth2):
    name = 'goclio'
    AUTHORIZATION_URL = 'https://app.goclio.com/oauth/authorize/'
    ACCESS_TOKEN_METHOD = 'POST'
    ACCESS_TOKEN_URL = 'https://app.goclio.com/oauth/token/'
    REDIRECT_STATE = False
    STATE_PARAMETER = False

    def get_user_details(self, response):
        """Return user details from GoClio account"""
        user = response.get('user', {})
        username = user.get('id', None)
        email = user.get('email', None)
        first_name, last_name = (user.get('first_name', None),
                                 user.get('last_name', None))
        fullname = '%s %s' % (first_name, last_name)

        return {'username': username,
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name,
                'email': email}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            'https://app.goclio.com/api/v2/users/who_am_i',
            params={'access_token': access_token}
        )

    def get_user_id(self, details, response):
        return response.get('user', {}).get('id')



"""
Trello OAuth1 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/trello.html
"""
from .oauth import BaseOAuth1


class TrelloOAuth(BaseOAuth1):

    """Trello OAuth authentication backend"""
    name = 'trello'
    ID_KEY = 'username'
    AUTHORIZATION_URL = 'https://trello.com/1/OAuthAuthorizeToken'
    REQUEST_TOKEN_URL = 'https://trello.com/1/OAuthGetRequestToken'
    ACCESS_TOKEN_URL = 'https://trello.com/1/OAuthGetAccessToken'

    EXTRA_DATA = [
        ('username', 'username'),
        ('email', 'email'),
        ('fullName', 'fullName')
    ]

    def get_user_details(self, response):
        """Return user details from Trello account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get('fullName')
        )
        return {'username': response.get('username'),
                'email': response.get('email'),
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token):
        """Return user data provided"""
        url = 'https://trello.com/1/members/me'
        try:
            return self.get_json(url, auth=self.oauth_auth(access_token))
        except ValueError:
            return None

    def auth_extra_arguments(self):
        return {
            'name': self.setting('APP_NAME', ''),
            # trello default expiration is '30days'
            'expiration': self.setting('EXPIRATION', 'never')
        }



import unittest2

from .oauth import OAuth1Test, OAuth2Test
from .open_id_connect import OpenIdConnectTestMixin


class ElixirOpenIdConnectTest(OpenIdConnectTestMixin, OAuth2Test):
    backend_path = 'social_core.backends.elixir.ElixirOpenIdConnect'
    issuer = 'https://login.elixir-czech.org/oidc/'
    openid_config_body = """
    {
        "claims_supported": [
            "sub",
            "name",
            "preferred_username",
            "given_name",
            "family_name",
            "middle_name",
            "nickname",
            "profile",
            "picture",
            "website",
            "gender",
            "zoneinfo",
            "locale",
            "updated_at",
            "birthdate",
            "email",
            "email_verified",
            "phone_number",
            "phone_number_verified",
            "address"
        ],
        "op_policy_uri": "https://login.elixir-czech.org/oidc/about",
        "subject_types_supported": [
            "public",
            "pairwise"
        ],
        "request_parameter_supported": true,
        "userinfo_signing_alg_values_supported": [
            "HS256",
            "HS384",
            "HS512",
            "RS256",
            "RS384",
            "RS512",
            "ES256",
            "ES384",
            "ES512",
            "PS256",
            "PS384",
            "PS512"
        ],
        "revocation_endpoint": "https://login.elixir-czech.org/oidc/revoke",
        "issuer": "https://login.elixir-czech.org/oidc/",
        "id_token_encryption_enc_values_supported": [
            "A256CBC+HS512",
            "A256GCM",
            "A192GCM",
            "A128GCM",
            "A128CBC-HS256",
            "A192CBC-HS384",
            "A256CBC-HS512",
            "A128CBC+HS256"
        ],
        "require_request_uri_registration": false,
        "grant_types_supported": [
            "authorization_code",
            "implicit",
            "urn:ietf:params:oauth:grant-type:jwt-bearer",
            "client_credentials",
            "urn:ietf:params:oauth:grant_type:redelegate",
            "urn:ietf:params:oauth:grant-type:device_code"
        ],
        "token_endpoint": "https://login.elixir-czech.org/oidc/token",
        "request_uri_parameter_supported": false,
        "service_documentation": "https://login.elixir-czech.org/oidc/about",
        "registration_endpoint": "https://login.elixir-czech.org/oidc/register",
        "jwks_uri": "https://login.elixir-czech.org/oidc/jwk",
        "userinfo_encryption_alg_values_supported": [
            "RSA-OAEP",
            "RSA-OAEP-256",
            "RSA1_5"
        ],
        "scopes_supported": [],
        "token_endpoint_auth_methods_supported": [
            "client_secret_post",
            "client_secret_basic",
            "client_secret_jwt",
            "private_key_jwt",
            "none"
        ],
        "userinfo_encryption_enc_values_supported": [
            "A256CBC+HS512",
            "A256GCM",
            "A192GCM",
            "A128GCM",
            "A128CBC-HS256",
            "A192CBC-HS384",
            "A256CBC-HS512",
            "A128CBC+HS256"
        ],
        "claim_types_supported": [
            "normal"
        ],
        "request_object_encryption_enc_values_supported": [
            "A256CBC+HS512",
            "A256GCM",
            "A192GCM",
            "A128GCM",
            "A128CBC-HS256",
            "A192CBC-HS384",
            "A256CBC-HS512",
            "A128CBC+HS256"
        ],
        "claims_parameter_supported": false,
        "id_token_encryption_alg_values_supported": [
            "RSA-OAEP",
            "RSA-OAEP-256",
            "RSA1_5"
        ],
        "code_challenge_methods_supported": [
            "plain",
            "S256"
        ],
        "token_endpoint_auth_signing_alg_values_supported": [
            "HS256",
            "HS384",
            "HS512",
            "RS256",
            "RS384",
            "RS512",
            "ES256",
            "ES384",
            "ES512",
            "PS256",
            "PS384",
            "PS512"
        ],
        "userinfo_endpoint": "https://login.elixir-czech.org/oidc/userinfo",
        "introspection_endpoint": "https://login.elixir-czech.org/oidc/introspect",
        "id_token_signing_alg_values_supported": [
            "HS256",
            "HS384",
            "HS512",
            "RS256",
            "RS384",
            "RS512",
            "ES256",
            "ES384",
            "ES512",
            "PS256",
            "PS384",
            "PS512",
            "none"
        ],
        "device_authorization_endpoint": "https://login.elixir-czech.org/oidc/devicecode",
        "op_tos_uri": "https://login.elixir-czech.org/oidc/about",
        "request_object_encryption_alg_values_supported": [
            "RSA-OAEP",
            "RSA-OAEP-256",
            "RSA1_5"
        ],
        "request_object_signing_alg_values_supported": [
            "HS256",
            "HS384",
            "HS512",
            "RS256",
            "RS384",
            "RS512",
            "ES256",
            "ES384",
            "ES512",
            "PS256",
            "PS384",
            "PS512"
        ],
        "response_types_supported": [
            "code",
            "token"
        ],
        "end_session_endpoint": "https://login.elixir-czech.org/oidc/endsession",
        "authorization_endpoint": "https://login.elixir-czech.org/oidc/authorize"
    }
    """



import time

from jwt import DecodeError, ExpiredSignature

from ..exceptions import AuthTokenError
from .oauth import BaseOAuth2

"""
OAuth2 Backend to work with microsoft graph.
"""


class MicrosoftOAuth2(BaseOAuth2):
    name = 'microsoft-graph'
    SCOPE_SEPARATOR = ' '
    AUTHORIZATION_URL = \
        'https://login.microsoftonline.com/common/oauth2/v2.0/authorize'
    ACCESS_TOKEN_URL = \
        'https://login.microsoftonline.com/common/oauth2/v2.0/token'

    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False
    DEFAULT_SCOPE = ['User.Read']

    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        self.process_error(self.data)
        state = self.validate_state()

        response = self.request_access_token(
            self.access_token_url(),
            data=self.auth_complete_params(state),
            headers=self.auth_headers(),
            auth=self.auth_complete_credentials(),
            method=self.ACCESS_TOKEN_METHOD
        )

        self.process_error(response)
        return self.do_auth(response['access_token'], response=response,
                            *args, **kwargs)

    def get_user_id(self, details, response):
        """Use user account id as unique id"""
        return response.get('id')

    def get_user_details(self, response):
        """Return user details from Microsoft online account"""
        email = response.get('mail')
        username = response.get('userPrincipalName')

        if '@' in username:
            if not email:
                email = username
            username = username.split('@', 1)[0]

        return {'username': username,
                'email': email,
                'fullname': response.get('displayName', ''),
                'first_name': response.get('givenName', ''),
                'last_name': response.get('surname', '')}

    def user_data(self, access_token, *args, **kwargs):
        """Return user data by querying Microsoft service"""
        try:
            return self.get_json(
                'https://graph.microsoft.com/v1.0/me',
                headers={
                    'Content-Type': 'application/x-www-form-urlencoded',
                    'Accept': 'application/json',
                    'Authorization': 'Bearer ' + access_token
                },
                method='GET'
            )
        except (DecodeError, ExpiredSignature) as error:
            raise AuthTokenError(self, error)

    def get_auth_token(self, user_id):
        """Return the access token for the given user, after ensuring that it
        has not expired, or refreshing it if so."""
        user = self.get_user(user_id=user_id)
        access_token = user.social_user.access_token
        expires_on = user.social_user.extra_data['expires_on']
        if expires_on <= int(time.time()):
            new_token_response = self.refresh_token(token=access_token)
            access_token = new_token_response['access_token']
        return access_token



import json
import unittest2

from .oauth import OAuth1Test, OAuth2Test
from .open_id_connect import OpenIdConnectTestMixin


class GlobusOpenIdConnectTest(OpenIdConnectTestMixin, OAuth2Test):
    backend_path = \
        'social_core.backends.globus.GlobusOpenIdConnect'
    issuer = 'https://auth.globus.org'
    openid_config_body = json.dumps({
        'issuer': 'https://auth.globus.org',
        'authorization_endpoint': 'https://auth.globus.org/v2/oauth2/authorize',
        'userinfo_endpoint': 'https://auth.globus.org/v2/oauth2/userinfo',
        'token_endpoint': 'https://auth.globus.org/v2/oauth2/token',
        'revocation_endpoint': 'https://auth.globus.org/v2/oauth2/token/revoke',
        'jwks_uri': 'https://auth.globus.org/jwk.json',
        'response_types_supported': [
            'code',
            'token',
            'token id_token',
            'id_token'
        ],
        'id_token_signing_alg_values_supported': [
            'RS512'
        ],
        'scopes_supported': [
            'openid',
            'email',
            'profile'
        ],
        'token_endpoint_auth_methods_supported': [
            'client_secret_basic'
        ],
        'claims_supported': [
            'at_hash',
            'aud',
            'email',
            'exp',
            'name',
            'nonce',
            'preferred_username',
            'iat',
            'iss',
            'sub'
        ],
        'subject_types_supported' : ['public']
    })



"""
Udata related backends.

Docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/udata.html
"""
from .oauth import BaseOAuth2


class UdataBaseOAuth2(BaseOAuth2):
    """Udata base OAuth authentication backend."""
    SCOPE_SEPARATOR = ','
    REDIRECT_STATE = False
    DEFAULT_SCOPE = ['default']
    ACCESS_TOKEN_METHOD = 'POST'

    def get_user_details(self, response):
        """Return user details from Udata account."""
        return {
            'username': response.get('first_name'),
            'email': response.get('email') or '',
            'first_name': response.get('first_name')
        }

    def user_data(self, access_token, *args, **kwargs):
        """Load user data from service."""
        return self.get_json(self.USER_DATA_URL, params={
            'access_token': access_token
        })


class DatagouvfrOAuth2(UdataBaseOAuth2):
    """Datagouvfr OAuth authentication backend."""
    name = 'datagouv'
    ACCESS_TOKEN_URL = 'https://www.data.gouv.fr/oauth/token'
    AUTHORIZATION_URL = 'https://www.data.gouv.fr/oauth/authorize'
    USER_DATA_URL = 'https://www.data.gouv.fr/api/1/me/'



import json

from httpretty import HTTPretty

from ...exceptions import AuthFailed
from .oauth import OAuth2Test


class ScistarterOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.scistarter.SciStarterOAuth2'
    user_data_url = 'https://scistarter.com/api/user_info'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'profile_id': 42006,
        'user_id': 5,
        'url': 'https://scistarter.com/user/foobar',
        'result': 'success',
        'handle': 'foobar',
        'email': 'foo@bar.com',
        'first_name': 'foo',
        'last_name': 'bar'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Shopify OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/shopify.html
"""
import imp
import six

from ..utils import handle_http_errors
from .oauth import BaseOAuth2
from ..exceptions import AuthFailed, AuthCanceled


class ShopifyOAuth2(BaseOAuth2):
    """Shopify OAuth2 authentication backend"""
    name = 'shopify'
    ID_KEY = 'shop'
    EXTRA_DATA = [
        ('shop', 'shop'),
        ('website', 'website'),
        ('expires', 'expires')
    ]
    REDIRECT_STATE = False

    @property
    def shopifyAPI(self):
        if not hasattr(self, '_shopify_api'):
            fp, pathname, description = imp.find_module('shopify')
            self._shopify_api = imp.load_module('shopify', fp, pathname,
                                                description)
        return self._shopify_api

    def get_user_details(self, response):
        """Use the shopify store name as the username"""
        return {
            'username': six.text_type(response.get('shop', '')).replace(
                '.myshopify.com', ''
            )
        }

    def extra_data(self, user, uid, response, details=None, *args, **kwargs):
        """Return access_token and extra defined names to store in
        extra_data field"""
        data = super(ShopifyOAuth2, self).extra_data(user, uid, response,
                                                     details, *args, **kwargs)
        session = self.shopifyAPI.Session(self.data.get('shop').strip())
        # Get, and store the permanent token
        token = session.request_token(data['access_token'])
        data['access_token'] = token
        return dict(data)

    def auth_url(self):
        key, secret = self.get_key_and_secret()
        self.shopifyAPI.Session.setup(api_key=key, secret=secret)
        scope = self.get_scope()
        state = self.state_token()
        self.strategy.session_set(self.name + '_state', state)
        redirect_uri = self.get_redirect_uri(state)
        session = self.shopifyAPI.Session(self.data.get('shop').strip())
        return session.create_permission_url(
            scope=scope,
            redirect_uri=redirect_uri
        )

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        self.process_error(self.data)
        access_token = None
        key, secret = self.get_key_and_secret()
        try:
            shop_url = self.data.get('shop')
            self.shopifyAPI.Session.setup(api_key=key, secret=secret)
            shopify_session = self.shopifyAPI.Session(shop_url, self.data)
            access_token = shopify_session.token
        except self.shopifyAPI.ValidationException:
            raise AuthCanceled(self)
        else:
            if not access_token:
                raise AuthFailed(self, 'Authentication Failed')
        return self.do_auth(access_token, shop_url, shopify_session.url,
                            *args, **kwargs)

    def do_auth(self, access_token, shop_url, website, *args, **kwargs):
        kwargs.update({
            'backend': self,
            'response': {
                'shop': shop_url,
                'website': 'http://{0}'.format(website),
                'access_token': access_token
            }
        })
        return self.strategy.authenticate(*args, **kwargs)



"""
Fitbit OAuth backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/fitbit.html
"""
import base64

from .oauth import BaseOAuth1, BaseOAuth2


class FitbitOAuth1(BaseOAuth1):
    """Fitbit OAuth1 authentication backend"""
    name = 'fitbit'
    AUTHORIZATION_URL = 'https://www.fitbit.com/oauth/authorize'
    REQUEST_TOKEN_URL = 'https://api.fitbit.com/oauth/request_token'
    ACCESS_TOKEN_URL = 'https://api.fitbit.com/oauth/access_token'
    ID_KEY = 'encodedId'
    EXTRA_DATA = [('encodedId', 'id'),
                  ('displayName', 'username')]

    def get_user_details(self, response):
        """Return user details from Fitbit account"""
        return {'username': response.get('displayName'),
                'email': ''}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            'https://api.fitbit.com/1/user/-/profile.json',
            auth=self.oauth_auth(access_token)
        )['user']


class FitbitOAuth2(BaseOAuth2):
    """Fitbit OAuth2 authentication backend"""
    name = 'fitbit'
    AUTHORIZATION_URL = 'https://www.fitbit.com/oauth2/authorize'
    ACCESS_TOKEN_URL = 'https://api.fitbit.com/oauth2/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REFRESH_TOKEN_URL = 'https://api.fitbit.com/oauth2/token'
    DEFAULT_SCOPE = ['profile']
    ID_KEY = 'encodedId'
    REDIRECT_STATE = False
    EXTRA_DATA = [('expires_in', 'expires'),
                  ('refresh_token', 'refresh_token', True),
                  ('encodedId', 'id'),
                  ('displayName', 'username')]

    def get_user_details(self, response):
        """Return user details from Fitbit account"""
        return {'username': response.get('displayName'),
                'email': ''}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        auth_header = {"Authorization": "Bearer %s" % access_token}
        return self.get_json(
            'https://api.fitbit.com/1/user/-/profile.json',
            headers=auth_header
        )['user']

    def auth_headers(self):
        tokens = '{0}:{1}'.format(*self.get_key_and_secret())
        tokens = base64.urlsafe_b64encode(tokens.encode())
        tokens = tokens.decode()
        return {
            'Authorization': 'Basic {0}'.format(tokens)
        }



import json

from ...exceptions import AuthAlreadyAssociated

from ..models import User
from .actions import BaseActionTest


class AssociateActionTest(BaseActionTest):
    expected_username = 'foobar'

    def setUp(self):
        super(AssociateActionTest, self).setUp()
        self.user = User(username='foobar', email='foo@bar.com')
        self.backend.strategy.session_set('username', self.user.username)

    def test_associate(self):
        self.do_login()
        self.assertTrue(len(self.user.social), 1)
        self.assertEqual(self.user.social[0].provider, 'github')

    def test_associate_with_partial_pipeline(self):
        self.do_login_with_partial_pipeline()
        self.assertEqual(len(self.user.social), 1)
        self.assertEqual(self.user.social[0].provider, 'github')


class MultipleAccountsTest(AssociateActionTest):
    alternative_user_data_body = json.dumps({
        'login': 'foobar2',
        'id': 2,
        'avatar_url': 'https://github.com/images/error/foobar2_happy.gif',
        'gravatar_id': 'somehexcode',
        'url': 'https://api.github.com/users/foobar2',
        'name': 'monalisa foobar2',
        'company': 'GitHub',
        'blog': 'https://github.com/blog',
        'location': 'San Francisco',
        'email': 'foo@bar.com',
        'hireable': False,
        'bio': 'There once was...',
        'public_repos': 2,
        'public_gists': 1,
        'followers': 20,
        'following': 0,
        'html_url': 'https://github.com/foobar2',
        'created_at': '2008-01-14T04:33:35Z',
        'type': 'User',
        'total_private_repos': 100,
        'owned_private_repos': 100,
        'private_gists': 81,
        'disk_usage': 10000,
        'collaborators': 8,
        'plan': {
            'name': 'Medium',
            'space': 400,
            'collaborators': 10,
            'private_repos': 20
        }
    })

    def test_multiple_social_accounts(self):
        self.do_login()
        self.do_login(user_data_body=self.alternative_user_data_body)
        self.assertEqual(len(self.user.social), 2)
        self.assertEqual(self.user.social[0].provider, 'github')
        self.assertEqual(self.user.social[1].provider, 'github')


class AlreadyAssociatedErrorTest(BaseActionTest):
    def setUp(self):
        super(AlreadyAssociatedErrorTest, self).setUp()
        self.user1 = User(username='foobar', email='foo@bar.com')
        self.user = None

    def tearDown(self):
        super(AlreadyAssociatedErrorTest, self).tearDown()
        self.user1 = None
        self.user = None

    def test_already_associated_error(self):
        self.user = self.user1
        self.do_login()
        self.user = User(username='foobar2', email='foo2@bar2.com')
        with self.assertRaisesRegexp(AuthAlreadyAssociated,
                                     'This account is already in use.'):
            self.do_login()



import json

from .oauth import OAuth2Test


class DeezerOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.deezer.DeezerOAuth2'
    user_data_url = 'http://api.deezer.com/user/me'
    expected_username = 'foobar'
    access_token_body = 'access_token=foobar&expires=0'
    user_data_body = json.dumps({
        'id': '1',
        'name': 'foobar',
        'lastname': '',
        'firstname': '',
        'status': 0,
        'birthday': '1970-01-01',
        'inscription_date': '2015-01-01',
        'gender': 'M',
        'link': 'https://www.deezer.com/profile/1',
        'picture': 'https://api.deezer.com/user/1/image',
        'picture_small': 'https://cdns-images.dzcdn.net/images/user//56x56-000000-80-0-0.jpg',
        'picture_medium': 'https://cdns-images.dzcdn.net/images/user//250x250-000000-80-0-0.jpg',
        'picture_big': 'https://cdns-images.dzcdn.net/images/user//500x500-000000-80-0-0.jpg',
        'country': 'FR',
        'lang': 'FR',
        'is_kid': False,
        'tracklist': 'https://api.deezer.com/user/1/flow',
        'type': 'user'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



# -*- coding: utf-8 -*-
import os
import sys
import json
import datetime
import unittest2
import base64
from calendar import timegm

from jose import jwt
from jose.jwk import RSAKey
from httpretty import HTTPretty

sys.path.insert(0, '..')

from ...exceptions import AuthTokenError


TEST_ROOT = os.path.dirname(os.path.dirname(__file__))

JWK_KEY = {
    'kty': 'RSA',
    'd': 'ZmswNokEvBcxW_Kvcy8mWUQOQCBdGbnM0xR7nhvGHC-Q24z3XAQWlMWbsmGc_R1o' \
         '_F3zK7DBlc3BokdRaO1KJirNmnHCw5TlnBlJrXiWpFBtVglUg98-4sRRO0VWnGXK' \
         'JPOkBQ6b_DYRO3b0o8CSpWowpiV6HB71cjXTqKPZf-aXU9WjCCAtxVjfIxgQFu5I' \
         '-G1Qah8mZeY8HK_y99L4f0siZcbUoaIcfeWBhxi14ODyuSAHt0sNEkhiIVBZE7QZ' \
         'm-SEP1ryT9VAaljbwHHPmg7NC26vtLZhvaBGbTTJnEH0ZubbN2PMzsfeNyoCIHy4' \
         '4QDSpQDCHfgcGOlHY_t5gQ',
    'e': 'AQAB',
    'use': 'sig',
    'kid': 'testkey',
    'alg': 'RS256',
    'n': 'pUfcJ8WFrVue98Ygzb6KEQXHBzi8HavCu8VENB2As943--bHPcQ-nScXnrRFAUg8' \
         'H5ZltuOcHWvsGw_AQifSLmOCSWJAPkdNb0w0QzY7Re8NrPjCsP58Tytp5LicF0Ao' \
         'Ag28UK3JioY9hXHGvdZsWR1Rp3I-Z3nRBP6HyO18pEgcZ91c9aAzsqu80An9X4DA' \
         'b1lExtZorvcd5yTBzZgr-MUeytVRni2lDNEpa6OFuopHXmg27Hn3oWAaQlbymd4g' \
         'ifc01oahcwl3ze2tMK6gJxa_TdCf1y99Yq6oilmVvZJ8kwWWnbPE-oDmOVPVnEyT' \
         'vYVCvN4rBT1DQ-x0F1mo2Q',
}

JWK_PUBLIC_KEY = {key: value for key, value in JWK_KEY.items() if key != 'd'}


class OpenIdConnectTestMixin(object):
    """
    Mixin to test OpenID Connect consumers. Inheriting classes should also
    inherit OAuth2Test.
    """
    client_key = 'a-key'
    client_secret = 'a-secret-key'
    issuer = None  # id_token issuer
    openid_config_body = None
    key = None

    def setUp(self):
        super(OpenIdConnectTestMixin, self).setUp()
        self.key = JWK_KEY.copy()
        self.public_key = JWK_PUBLIC_KEY.copy()

        HTTPretty.register_uri(HTTPretty.GET,
          self.backend.OIDC_ENDPOINT + '/.well-known/openid-configuration',
          status=200,
          body=self.openid_config_body
        )
        oidc_config = json.loads(self.openid_config_body)

        def jwks(_request, _uri, headers):
            return 200, headers, json.dumps({'keys': [self.key]})

        HTTPretty.register_uri(HTTPretty.GET,
                               oidc_config.get('jwks_uri'),
                               status=200,
                               body=json.dumps({'keys': [self.public_key]}))

    def extra_settings(self):
        settings = super(OpenIdConnectTestMixin, self).extra_settings()
        settings.update({
            'SOCIAL_AUTH_{0}_KEY'.format(self.name): self.client_key,
            'SOCIAL_AUTH_{0}_SECRET'.format(self.name): self.client_secret,
            'SOCIAL_AUTH_{0}_ID_TOKEN_DECRYPTION_KEY'.format(self.name):
                self.client_secret
        })
        return settings

    def get_id_token(self, client_key=None, expiration_datetime=None,
                     issue_datetime=None, nonce=None, issuer=None):
        """
        Return the id_token to be added to the access token body.
        """
        return {
            'iss': issuer,
            'nonce': nonce,
            'aud': client_key,
            'azp': client_key,
            'exp': expiration_datetime,
            'iat': issue_datetime,
            'sub': '1234'
        }

    def prepare_access_token_body(self, client_key=None, tamper_message=False,
                                  expiration_datetime=None,
                                  issue_datetime=None, nonce=None,
                                  issuer=None):
        """
        Prepares a provider access token response. Arguments:

        client_id       -- (str) OAuth ID for the client that requested
                                 authentication.
        expiration_time -- (datetime) Date and time after which the response
                                      should be considered invalid.
        """

        body = {'access_token': 'foobar', 'token_type': 'bearer'}
        client_key = client_key or self.client_key
        now = datetime.datetime.utcnow()
        expiration_datetime = expiration_datetime or \
                              (now + datetime.timedelta(seconds=30))
        issue_datetime = issue_datetime or now
        nonce = nonce or 'a-nonce'
        issuer = issuer or self.issuer
        id_token = self.get_id_token(
            client_key,
            timegm(expiration_datetime.utctimetuple()),
            timegm(issue_datetime.utctimetuple()),
            nonce,
            issuer
        )

        body['id_token'] = jwt.encode(
            id_token,
            key=dict(self.key,
                     iat=timegm(issue_datetime.utctimetuple()),
                     nonce=nonce),
            algorithm='RS256',
            access_token='foobar'
        )

        if tamper_message:
            header, msg, sig = body['id_token'].split('.')
            id_token['sub'] = '1235'
            msg = base64.encodestring(json.dumps(id_token).encode()).decode()
            body['id_token'] = '.'.join([header, msg, sig])

        return json.dumps(body)

    def authtoken_raised(self, expected_message, **access_token_kwargs):
        self.access_token_body = self.prepare_access_token_body(
            **access_token_kwargs
        )
        with self.assertRaisesRegexp(AuthTokenError, expected_message):
            self.do_login()

    def test_invalid_signature(self):
        self.authtoken_raised(
            'Token error: Signature verification failed',
            tamper_message=True
        )

    def test_expired_signature(self):
        expiration_datetime = datetime.datetime.utcnow() - \
                              datetime.timedelta(seconds=30)
        self.authtoken_raised('Token error: Signature has expired',
                              expiration_datetime=expiration_datetime)

    def test_invalid_issuer(self):
        self.authtoken_raised('Token error: Invalid issuer',
                              issuer='someone-else')

    def test_invalid_audience(self):
        self.authtoken_raised('Token error: Invalid audience',
                              client_key='someone-else')

    def test_invalid_issue_time(self):
        expiration_datetime = datetime.datetime.utcnow() - \
                              datetime.timedelta(hours=1)
        self.authtoken_raised('Token error: Incorrect id_token: iat',
                              issue_datetime=expiration_datetime)

    def test_invalid_nonce(self):
        self.authtoken_raised(
            'Token error: Incorrect id_token: nonce',
            nonce='something-wrong'
        )



import json

from .oauth import OAuth2Test


class UniverseAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.universe.UniverseOAuth2'
    user_data_url = 'https://www.universe.com/api/v2/current_user'
    expected_username = 'scott+awesome@universe.com'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps(
        {
            'current_user': {
                'id': '123456',
                'slug': 'scott-vitale',
                'first_name': 'Scott',
                'last_name': 'Vitale',
                'created_at': '2019-01-08T15:49:42.514Z',
                'updated_at': '2019-01-17T19:41:39.711Z',
                'email': 'scott+awesome@universe.com',
            }
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



from ...utils import PARTIAL_TOKEN_SESSION_NAME
from ..models import User
from .actions import BaseActionTest


class LoginActionTest(BaseActionTest):
    def test_login(self):
        self.do_login()

    def test_login_with_partial_pipeline(self):
        self.do_login_with_partial_pipeline()

    def test_fields_stored_in_session(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_FIELDS_STORED_IN_SESSION': ['foo', 'bar']
        })
        self.strategy.set_request_data({'foo': '1', 'bar': '2'}, self.backend)
        self.do_login()
        self.assertEqual(self.strategy.session_get('foo'), '1')
        self.assertEqual(self.strategy.session_get('bar'), '2')

    def test_redirect_value(self):
        self.strategy.set_request_data({'next': '/after-login'}, self.backend)
        redirect = self.do_login(after_complete_checks=False)
        self.assertEqual(redirect.url, '/after-login')

    def test_login_with_invalid_partial_pipeline(self):
        def before_complete():
            partial_token = self.strategy.session_get(
                PARTIAL_TOKEN_SESSION_NAME
            )
            partial = self.strategy.storage.partial.load(partial_token)
            partial.data['backend'] = 'foobar'
        self.do_login_with_partial_pipeline(before_complete)

    def test_new_user(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_NEW_USER_REDIRECT_URL': '/new-user'
        })
        redirect = self.do_login(after_complete_checks=False)
        self.assertEqual(redirect.url, '/new-user')

    def test_inactive_user(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_INACTIVE_USER_URL': '/inactive'
        })
        User.set_active(False)
        redirect = self.do_login(after_complete_checks=False)
        self.assertEqual(redirect.url, '/inactive')

    def test_invalid_user(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_LOGIN_ERROR_URL': '/error',
            'SOCIAL_AUTH_PIPELINE': (
                'social_core.pipeline.social_auth.social_details',
                'social_core.pipeline.social_auth.social_uid',
                'social_core.pipeline.social_auth.auth_allowed',
                'social_core.pipeline.social_auth.social_user',
                'social_core.pipeline.user.get_username',
                'social_core.pipeline.user.create_user',
                'social_core.pipeline.social_auth.associate_user',
                'social_core.pipeline.social_auth.load_extra_data',
                'social_core.pipeline.user.user_details',
                'social_core.tests.pipeline.remove_user'
            )
        })
        redirect = self.do_login(after_complete_checks=False)
        self.assertEqual(redirect.url, '/error')



import json

from .oauth import OAuth2Test


class ChatworkOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.chatwork.ChatworkOAuth2'
    user_data_url = 'https://api.chatwork.com/v2/me'
    expected_username = 'hogehoge'
    access_token_body = json.dumps({
        'access_token': 'pyopyopyopyopyopyopyopyopyopyo',
        'token_type': 'Bearer',
        'expires_in': '1501138041000',
        'refresh_token': 'pyopyopyopyopyopyo',
        'scope': 'rooms.all:read_write'
    })

    user_data_body = json.dumps({
        'account_id': 123,
        'room_id': 322,
        'name': 'Foo Bar',
        'chatwork_id': 'hogehoge',
        'organization_id': 101,
        'organization_name': 'Foo foobar',
        'department': 'Support',
        'title': 'CMO',
        'url': 'http://www.example.com',
        'introduction': '',
        'mail': 'hogehoge@example.com',
        'tel_organization': '',
        'tel_extension': '',
        'tel_mobile': '',
        'skype': '',
        'facebook': '',
        'twitter': '',
        'avatar_image_url': 'https://www.example.com/hogehoge.jpg',
        'login_mail': 'hogehoge@example.com'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Google OpenId, OAuth2, OAuth1, Google+ Sign-in backends, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/google.html
"""
from ..utils import handle_http_errors
from .open_id import OpenIdAuth
from .oauth import BaseOAuth2, BaseOAuth1
from ..exceptions import AuthMissingParameter


class BaseGoogleAuth(object):
    def get_user_id(self, details, response):
        """Use google email as unique id"""
        if self.setting('USE_UNIQUE_USER_ID', False):
            if 'sub' in response:
                return response['sub']
            else:
                return response['id']
        else:
            return details['email']

    def get_user_details(self, response):
        """Return user details from Google API account"""
        if 'email' in response:
            email = response['email']
        else:
            email = ''

        name, given_name, family_name = (
            response.get('name', ''),
            response.get('given_name', ''),
            response.get('family_name', ''),
        )

        fullname, first_name, last_name = self.get_user_names(
            name, given_name, family_name
        )
        return {'username': email.split('@', 1)[0],
                'email': email,
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}


class BaseGoogleOAuth2API(BaseGoogleAuth):
    def user_data(self, access_token, *args, **kwargs):
        """Return user data from Google API"""
        return self.get_json(
            'https://www.googleapis.com/oauth2/v3/userinfo',
            headers={
                'Authorization': 'Bearer %s' % access_token,
            },
        )

    def revoke_token_params(self, token, uid):
        return {'token': token}

    def revoke_token_headers(self, token, uid):
        return {'Content-type': 'application/json'}


class GoogleOAuth2(BaseGoogleOAuth2API, BaseOAuth2):
    """Google OAuth2 authentication backend"""
    name = 'google-oauth2'
    REDIRECT_STATE = False
    AUTHORIZATION_URL = 'https://accounts.google.com/o/oauth2/auth'
    ACCESS_TOKEN_URL = 'https://accounts.google.com/o/oauth2/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REVOKE_TOKEN_URL = 'https://accounts.google.com/o/oauth2/revoke'
    REVOKE_TOKEN_METHOD = 'GET'
    # The order of the default scope is important
    DEFAULT_SCOPE = ['openid', 'email', 'profile']
    EXTRA_DATA = [
        ('refresh_token', 'refresh_token', True),
        ('expires_in', 'expires'),
        ('token_type', 'token_type', True)
    ]


class GooglePlusAuth(BaseGoogleOAuth2API, BaseOAuth2):
    name = 'google-plus'
    REDIRECT_STATE = False
    STATE_PARAMETER = False
    AUTHORIZATION_URL = 'https://accounts.google.com/o/oauth2/auth'
    ACCESS_TOKEN_URL = 'https://accounts.google.com/o/oauth2/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REVOKE_TOKEN_URL = 'https://accounts.google.com/o/oauth2/revoke'
    REVOKE_TOKEN_METHOD = 'GET'
    DEFAULT_SCOPE = [
        'https://www.googleapis.com/auth/plus.login',
        'https://www.googleapis.com/auth/plus.me',
    ]
    EXTRA_DATA = [
        ('id', 'user_id'),
        ('refresh_token', 'refresh_token', True),
        ('expires_in', 'expires'),
        ('access_type', 'access_type', True),
        ('code', 'code')
    ]

    def auth_complete_params(self, state=None):
        params = super(GooglePlusAuth, self).auth_complete_params(state)
        if self.data.get('access_token'):
            # Don't add postmessage if this is plain server-side workflow
            params['redirect_uri'] = 'postmessage'
        return params

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        if 'access_token' in self.data:  # Client-side workflow
            token = self.data.get('access_token')
            response = self.get_json(
                'https://www.googleapis.com/oauth2/v3/tokeninfo',
                params={'access_token': token}
            )
            self.process_error(response)
            return self.do_auth(token, response=response, *args, **kwargs)
        elif 'code' in self.data:  # Server-side workflow
            response = self.request_access_token(
                self.ACCESS_TOKEN_URL,
                data=self.auth_complete_params(),
                headers=self.auth_headers(),
                method=self.ACCESS_TOKEN_METHOD
            )
            self.process_error(response)
            return self.do_auth(response['access_token'],
                                response=response,
                                *args, **kwargs)
        elif 'id_token' in self.data:  # Client-side workflow
            token = self.data.get('id_token')
            return self.do_auth(token, *args, **kwargs)
        else:
            raise AuthMissingParameter(self, 'access_token, id_token, or code')

    def user_data(self, access_token, *args, **kwargs):
        if 'id_token' not in self.data:
            return super(GooglePlusAuth, self).user_data(access_token, *args,
                                                         **kwargs)
        response = self.get_json(
            'https://www.googleapis.com/oauth2/v3/tokeninfo',
            params={'id_token': access_token}
        )
        self.process_error(response)
        return response


class GoogleOAuth(BaseGoogleAuth, BaseOAuth1):
    """Google OAuth authorization mechanism"""
    name = 'google-oauth'
    AUTHORIZATION_URL = 'https://www.google.com/accounts/OAuthAuthorizeToken'
    REQUEST_TOKEN_URL = 'https://www.google.com/accounts/OAuthGetRequestToken'
    ACCESS_TOKEN_URL = 'https://www.google.com/accounts/OAuthGetAccessToken'
    DEFAULT_SCOPE = ['https://www.googleapis.com/auth/userinfo#email']

    def user_data(self, access_token, *args, **kwargs):
        """Return user data from Google API"""
        return self.get_querystring(
            'https://www.googleapis.com/userinfo/email',
            auth=self.oauth_auth(access_token)
        )

    def get_key_and_secret(self):
        """Return Google OAuth Consumer Key and Consumer Secret pair, uses
        anonymous by default, beware that this marks the application as not
        registered and a security badge is displayed on authorization page.
        http://code.google.com/apis/accounts/docs/OAuth_ref.html#SigningOAuth
        """
        key_secret = super(GoogleOAuth, self).get_key_and_secret()
        if key_secret == (None, None):
            key_secret = ('anonymous', 'anonymous')
        return key_secret


class GoogleOpenId(OpenIdAuth):
    name = 'google'
    URL = 'https://www.google.com/accounts/o8/id'

    def get_user_id(self, details, response):
        """
        Return user unique id provided by service. For google user email
        is unique enought to flag a single user. Email comes from schema:
        http://axschema.org/contact/email
        """
        return details['email']



import json

from .oauth import OAuth2Test


class SlackOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.slack.SlackOAuth2'
    user_data_url = 'https://slack.com/api/users.identity'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'ok': True,
        'user': {
            'email': 'foobar@example.com',
            'name': 'Foo Bar',
            'id': u'123456'
        },
        'team': {
            'id': u'456789'
        },
        'scope': u'identity.basic,identity.email'
    })
    expected_username = 'foobar'

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()


class SlackOAuth2TestTeamName(SlackOAuth2Test):
    expected_username = 'foobar@Square'
    user_data_body = json.dumps({
        'ok': True,
        'user': {
            'email': 'foobar@example.com',
            'name': 'Foo Bar',
            'id': u'123456'
        },
        'team': {
            'id': u'456789',
            'name': u'Square',
        },
        'scope': u'identity.basic,identity.email,identity.team'
    })


class SlackOAuth2TestUnicodeTeamName(SlackOAuth2Test):
    user_data_body = json.dumps({
        'ok': True,
        'user': {
            'email': 'foobar@example.com',
            'name': 'Foo Bar',
            'id': u'123456'
        },
        'team': {
            'id': u'456789',
            'name': u'Square \u221a team',
        },
        'scope': u'identity.basic,identity.email,identity.team'
    })

    def test_login(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_SLACK_USERNAME_WITH_TEAM': False
        })
        self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_SLACK_USERNAME_WITH_TEAM': False
        })
        self.do_partial_pipeline()



"""
Orbi OAuth2 backend
"""
from .oauth import BaseOAuth2


class OrbiOAuth2(BaseOAuth2):
    """Orbi OAuth2 authentication backend"""
    name = 'orbi'
    AUTHORIZATION_URL = 'https://login.orbi.kr/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://login.orbi.kr/oauth/token'
    ACCESS_TOKEN_METHOD = 'POST'
    EXTRA_DATA = [
        ('imin', 'imin'),
        ('nick', 'nick'),
        ('photo', 'photo'),
        ('sex', 'sex'),
        ('birth', 'birth'),
    ]

    def get_user_id(self, details, response):
        return response.get('id')

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(
            response.get('name', ''),
            response.get('first_name', ''),
            response.get('last_name', '')
        )
        return {
            'username': response.get('username', response.get('name')),
            'email': response.get('email', ''),
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Load user data from orbi"""
        return self.get_json('https://login.orbi.kr/oauth/user/get', params={
            'access_token': access_token
        })



import json
import time
import unittest

import jwt

from .oauth import OAuth2Test


_PRIVATE_KEY_HEADERLESS = '''
MIIEowIBAAKCAQEAvyo2hx1L3ALHeUd/6xk/lIhTyZ/HJZ+Sss/ge6T6gPdES4Dw
BvwGlAp21iEbmjmizsv6+ZsyuKZUiC1J4A90lmIA57aYXHHoh9GBWQZzXeCNgghP
JpGYYSCN+1qeD4nbwD9cQBtPrGBpPpPtv2a/xdPqDm5ko6adMhmbm8e4Me/ppWPi
0U+skWQJepBhjEt3x+AOMKDv2TUBWOc3mYFNkr9qNOPe7FxnqUk6ZtkI3QNjZTky
AU7cat87u1vT5thAxVY18i1GfSZwtQbU3Ba6hXI5SIHB1lS88SJ9/+E/flJJPD2N
Nzv2z3HAVuTUOYi48fnXFHpJLGv+mGLNtE77hwIDAQABAoIBAQCUyQYno2Wnl4Ip
orys/rm9oV2VUAZwAgLrqV/O3Fkch1dgbLpktUNpdbuIbbxODQ3qZliwbVrM3Khu
VNFq0pyrbxvFPRjY2s9g5m8GGz8vkdaRnmX8XtV6wxu+xoi/D006FBZ4zsj0IRXI
3tnsXsxj7Mv+72zk8ojmtYend4qlUfzBVTpMRDc4XDC4Ya91fgFgfibtUE1qc8Ap
ctCzk9wZgN9SOKXHcKANhqC8BmQv7NspI0RT0Oq0n/U921P/+y1M++Z0Z2vQvjiR
GvDpSlnmlLB3S3E7zHbmksyUyK4Ab7xYi51yFKgrYjtaM3QLCTlSmTQBM8EhjZNG
VJvgEI4BAoGBAOQzd4PdB3jW6OnUXDrzWqfCAgpMxWtAmP5h+LrmsrV+upXggRnv
iNzTqiq0QpdrgYrikFGDkGEsqfrnd7IkskDMT8PMuncmDkuF67kg6G4/vzyz5QyA
jolf6qmHQPMCfxtCPUZZOUcuRKnj5KVIGfJu2gu5Z2lEysuV5ZQE0OiNAoGBANZz
wqLhqN376MT3YEbsOcWXYBDXX+FzWAYHsxf4APHJrNr6pkM4dTsvXU2tQIUV4N3c
SJvMVSI91VL8mdgQxHKaECUizye19brJ1BysbeFjBCK+Y6bYMd8n0Hdxxi/ZyeF9
AzfIPQN2uZSzU/I+Nt0tlz/SCoL4Qi3FQpgtLoFjAoGAb8XVsDy+wC1jf8SIOEei
C7E3FpxrxhCp309VaRY+Si98bJS+J1nwC1mRa8FHLKt3k/NNBOAQA8jAqShetF7N
AHgSSbEpU9rL/anmv5Kixf1rSexDMFB3gEn+wnKBGYYLg+p54M8rAvZio2QARgR+
0QQCwONbB3Cuc/FDtbB2MrECgYAhSqtGmf2bKIZEPZsGp5l4YT2an7TUzRE3Lm7R
I8ERyBs7i3nQKa2ZWIsFigXgIztbdd0Xwqrcu/in/2rqrf+xQtWKzlKWeZsCOl7h
bKtKOBLmSeQyfJGRcR7dzB3WQ9shVETxnfZK2V2KBiTcEGh4AaHfWH4lQuETNfJW
qXz0vQKBgDVz+ZvULA/OZWXrOI1il7KoahWdb9vr8VhWgHKnDW7hInDFh6SEQHn6
mSNns0AssDwr4TheET7klb7AvbBKrNSP/Tz9AzkwMz148T2ffkPFMZRuvRT+eQ5Z
ey4gIBKESJF6X9fefiawCrI3+PC7x9x0ngP9R4t/OzDWVAYn9gmd
'''.strip()

_PRIVATE_KEY = '''
-----BEGIN RSA PRIVATE KEY-----
{_PRIVATE_KEY_HEADERLESS}
-----END RSA PRIVATE KEY-----
'''.format(
    _PRIVATE_KEY_HEADERLESS=_PRIVATE_KEY_HEADERLESS
)

_PUBLIC_KEY_HEADERLESS = '''
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAvyo2hx1L3ALHeUd/6xk/
lIhTyZ/HJZ+Sss/ge6T6gPdES4DwBvwGlAp21iEbmjmizsv6+ZsyuKZUiC1J4A90
lmIA57aYXHHoh9GBWQZzXeCNgghPJpGYYSCN+1qeD4nbwD9cQBtPrGBpPpPtv2a/
xdPqDm5ko6adMhmbm8e4Me/ppWPi0U+skWQJepBhjEt3x+AOMKDv2TUBWOc3mYFN
kr9qNOPe7FxnqUk6ZtkI3QNjZTkyAU7cat87u1vT5thAxVY18i1GfSZwtQbU3Ba6
hXI5SIHB1lS88SJ9/+E/flJJPD2NNzv2z3HAVuTUOYi48fnXFHpJLGv+mGLNtE77
hwIDAQAB
'''.strip()

_PUBLIC_KEY = '''
-----BEGIN PUBLIC KEY-----
{_PUBLIC_KEY_HEADERLESS}
-----END PUBLIC KEY-----
'''.format(
    _PUBLIC_KEY_HEADERLESS=_PUBLIC_KEY_HEADERLESS
)

_KEY = 'example'
_SECRET = '1234abcd-1234-abcd-1234-abcd1234adcd'

_AUTHORIZATION_URL = 'https://sso.example.com/auth/realms/example/protocol/openid-connect/auth'
_ACCESS_TOKEN_URL = 'https://sso.example.com/auth/realms/example/protocol/openid-connect/token'

_ALGORITHM = 'RS256'
_AUTH_TIME = int(time.time())
_PAYLOAD = {
    'preferred_username': 'john.doe',
    'email': 'john.doe@example.com',
    'name': 'John Doe',
    'given_name': 'John',
    'family_name': 'Doe',

    'iss': 'https://sso.example.com',
    'sub': 'john.doe',
    'aud': _KEY,
    'exp': _AUTH_TIME + 3600,
    'iat': _AUTH_TIME,
}


def _encode(
    payload,
    key=_PRIVATE_KEY,
    algorithm=_ALGORITHM
):
    return jwt.encode(payload, key=key, algorithm=algorithm).decode('utf-8')


def _decode(
    token,
    key=_PUBLIC_KEY,
    algorithms=_ALGORITHM,
    audience=_KEY,
):
    return jwt.decode(token, key=key, algorithms=algorithms, audience=audience)


class KeycloakOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.keycloak.KeycloakOAuth2'
    expected_username = 'john.doe'
    access_token_body = json.dumps({
        'token_type': 'Bearer',
        'id_token': _encode(_PAYLOAD),
        'access_token': _encode(_PAYLOAD),
    })

    def extra_settings(self):
        return {
            'SOCIAL_AUTH_KEYCLOAK_KEY': _KEY,
            'SOCIAL_AUTH_KEYCLOAK_SECRET': _SECRET,
            'SOCIAL_AUTH_KEYCLOAK_PUBLIC_KEY': _PUBLIC_KEY_HEADERLESS,
            'SOCIAL_AUTH_KEYCLOAK_ALGORITHM': _ALGORITHM,
            'SOCIAL_AUTH_KEYCLOAK_AUTHORIZATION_URL': _AUTHORIZATION_URL,
            'SOCIAL_AUTH_KEYCLOAK_ACCESS_TOKEN_URL': _ACCESS_TOKEN_URL,
        }

    def test_encode_decode(self):
        token = _encode(_PAYLOAD)
        self.assertEqual(_PAYLOAD, _decode(token))

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Backend for OpenID Connect ELIXIR AAI
https://www.elixir-europe.org/services/compute/aai
"""

from social_core.backends.open_id_connect import OpenIdConnectAuth


class ElixirOpenIdConnect(OpenIdConnectAuth):
    name = 'elixir'
    OIDC_ENDPOINT = 'https://login.elixir-czech.org/oidc'
    EXTRA_DATA = [
        ('expires_in', 'expires_in', True),
        ('refresh_token', 'refresh_token', True),
        ('id_token', 'id_token', True),
        ('other_tokens', 'other_tokens', True),
    ]

    def get_user_details(self, response):
        username_key = self.setting('USERNAME_KEY', default=self.USERNAME_KEY)
        name = response.get('name') or ''
        fullname, first_name, last_name = self.get_user_names(name)
        return {'username': response.get(username_key),
                'email': response.get('email'),
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}



"""
Openshift OAuth2 backend
"""
import requests

from six.moves.urllib.parse import urljoin

from ..utils import append_slash
from .oauth import BaseOAuth2


class OpenshiftOAuth2(BaseOAuth2):
    name = 'openshift'
    ACCESS_TOKEN_METHOD = 'POST'

    def access_token_url(self):
        return urljoin(append_slash(self.setting('URL')), 'oauth/token')

    def authorization_url(self):
        return urljoin(append_slash(self.setting('URL')), 'oauth/authorize')

    def get_user_id(self, details, response):
        return response['metadata']['uid']

    def get_user_details(self, response):
        """Return user details from openshift account"""
        username = response['metadata']['name']
        email = response['metadata']['name']
        return {'username': username,
                'email': email}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        headers = {'Authorization': 'Bearer ' + access_token}

        return requests.get(
            urljoin(append_slash(self.setting('URL')), 'oapi/v1/users/~'),
            headers=headers
        ).json()



import json

from .oauth import OAuth2Test


class ItembaseOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.itembase.ItembaseOAuth2'
    user_data_url = 'https://users.itembase.com/v1/me'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        "access_token": "foobar-token",
        "expires_in": 2592000,
        "token_type": "bearer",
        "scope": "user.minimal",
        "refresh_token": "foobar-refresh-token"
    })
    user_data_body = json.dumps({
        "uuid": "a4b91ee7-ec1a-49b9-afce-371dc8797749",
        "username": "foobar",
        "email": "foobar@itembase.biz",
        "first_name": "Foo",
        "middle_name": None,
        "last_name": "Bar",
        "name_format": "first middle last",
        "locale": "en",
        "preferred_currency": "EUR"
    })
    refresh_token_body = json.dumps({
        "access_token": "foobar-new-token",
        "expires_in": 2592000,
        "token_type": "bearer",
        "scope": "user.minimal",
        "refresh_token": "foobar-new-refresh-token"
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()


class ItembaseOAuth2SandboxTest(OAuth2Test):
    backend_path = 'social_core.backends.itembase.ItembaseOAuth2Sandbox'
    user_data_url = 'http://sandbox.users.itembase.io/v1/me'



"""
Mendeley OAuth1 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/mendeley.html
"""
from .oauth import BaseOAuth1, BaseOAuth2


class MendeleyMixin(object):
    SCOPE_SEPARATOR = '+'
    EXTRA_DATA = [('profile_id', 'profile_id'),
                  ('name', 'name'),
                  ('bio', 'bio')]

    def get_user_id(self, details, response):
        return response['id']

    def get_user_details(self, response):
        """Return user details from Mendeley account"""
        profile_id = response['id']
        name = response['display_name']
        bio = response['link']
        return {'profile_id': profile_id,
                'name': name,
                'bio': bio}

    def user_data(self, access_token, *args, **kwargs):
        """Return user data provided"""
        values = self.get_user_data(access_token)
        values.update(values)
        return values

    def get_user_data(self, access_token):
        raise NotImplementedError('Implement in subclass')


class MendeleyOAuth(MendeleyMixin, BaseOAuth1):
    name = 'mendeley'
    AUTHORIZATION_URL = 'http://api.mendeley.com/oauth/authorize/'
    REQUEST_TOKEN_URL = 'http://api.mendeley.com/oauth/request_token/'
    ACCESS_TOKEN_URL = 'http://api.mendeley.com/oauth/access_token/'

    def get_user_data(self, access_token):
        return self.get_json(
            'http://api.mendeley.com/oapi/profiles/info/me/',
            auth=self.oauth_auth(access_token)
        )


class MendeleyOAuth2(MendeleyMixin, BaseOAuth2):
    name = 'mendeley-oauth2'
    AUTHORIZATION_URL = 'https://api-oauth2.mendeley.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://api-oauth2.mendeley.com/oauth/token'
    ACCESS_TOKEN_METHOD = 'POST'
    DEFAULT_SCOPE = ['all']
    REDIRECT_STATE = False
    EXTRA_DATA = MendeleyMixin.EXTRA_DATA + [
        ('refresh_token', 'refresh_token'),
        ('expires_in', 'expires_in'),
        ('token_type', 'token_type'),
    ]

    def get_user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            'https://api.mendeley.com/profiles/me/',
            headers={'Authorization': 'Bearer {0}'.format(access_token)}
        )



import json

from ..utils import PARTIAL_TOKEN_SESSION_NAME
from ..exceptions import AuthException

from .models import TestUserSocialAuth, TestStorage, User
from .strategy import TestStrategy
from .actions.actions import BaseActionTest


class IntegrityError(Exception):
    pass


class UnknownError(Exception):
    pass


class IntegrityErrorUserSocialAuth(TestUserSocialAuth):
    @classmethod
    def create_social_auth(cls, user, uid, provider):
        raise IntegrityError()

    @classmethod
    def get_social_auth(cls, provider, uid):
        if not hasattr(cls, '_called_times'):
            cls._called_times = 0
        cls._called_times += 1
        if cls._called_times == 2:
            user = list(User.cache.values())[0]
            return IntegrityErrorUserSocialAuth(user, provider, uid)
        else:
            return super(IntegrityErrorUserSocialAuth, cls).get_social_auth(
                provider, uid
            )


class IntegrityErrorStorage(TestStorage):
    user = IntegrityErrorUserSocialAuth

    @classmethod
    def is_integrity_error(cls, exception):
        """Check if given exception flags an integrity error in the DB"""
        return isinstance(exception, IntegrityError)


class UnknownErrorUserSocialAuth(TestUserSocialAuth):
    @classmethod
    def create_social_auth(cls, user, uid, provider):
        raise UnknownError()


class UnknownErrorStorage(IntegrityErrorStorage):
    user = UnknownErrorUserSocialAuth


class IntegrityErrorOnLoginTest(BaseActionTest):
    def setUp(self):
        self.strategy = TestStrategy(IntegrityErrorStorage)
        super(IntegrityErrorOnLoginTest, self).setUp()

    def test_integrity_error(self):
        self.do_login()


class UnknownErrorOnLoginTest(BaseActionTest):
    def setUp(self):
        self.strategy = TestStrategy(UnknownErrorStorage)
        super(UnknownErrorOnLoginTest, self).setUp()

    def test_unknown_error(self):
        with self.assertRaises(UnknownError):
            self.do_login()


class EmailAsUsernameTest(BaseActionTest):
    expected_username = 'foo@bar.com'

    def test_email_as_username(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_USERNAME_IS_FULL_EMAIL': True
        })
        self.do_login()


class RandomUsernameTest(BaseActionTest):
    user_data_body = json.dumps({
        'id': 1,
        'avatar_url': 'https://github.com/images/error/foobar_happy.gif',
        'gravatar_id': 'somehexcode',
        'url': 'https://api.github.com/users/foobar',
        'name': 'monalisa foobar',
        'company': 'GitHub',
        'blog': 'https://github.com/blog',
        'location': 'San Francisco',
        'email': 'foo@bar.com',
        'hireable': False,
        'bio': 'There once was...',
        'public_repos': 2,
        'public_gists': 1,
        'followers': 20,
        'following': 0,
        'html_url': 'https://github.com/foobar',
        'created_at': '2008-01-14T04:33:35Z',
        'type': 'User',
        'total_private_repos': 100,
        'owned_private_repos': 100,
        'private_gists': 81,
        'disk_usage': 10000,
        'collaborators': 8,
        'plan': {
            'name': 'Medium',
            'space': 400,
            'collaborators': 10,
            'private_repos': 20
        }
    })

    def test_random_username(self):
        self.do_login(after_complete_checks=False)


class SluggedUsernameTest(BaseActionTest):
    expected_username = 'foo-bar'
    user_data_body = json.dumps({
        'login': 'Foo Bar',
        'id': 1,
        'avatar_url': 'https://github.com/images/error/foobar_happy.gif',
        'gravatar_id': 'somehexcode',
        'url': 'https://api.github.com/users/foobar',
        'name': 'monalisa foobar',
        'company': 'GitHub',
        'blog': 'https://github.com/blog',
        'location': 'San Francisco',
        'email': 'foo@bar.com',
        'hireable': False,
        'bio': 'There once was...',
        'public_repos': 2,
        'public_gists': 1,
        'followers': 20,
        'following': 0,
        'html_url': 'https://github.com/foobar',
        'created_at': '2008-01-14T04:33:35Z',
        'type': 'User',
        'total_private_repos': 100,
        'owned_private_repos': 100,
        'private_gists': 81,
        'disk_usage': 10000,
        'collaborators': 8,
        'plan': {
            'name': 'Medium',
            'space': 400,
            'collaborators': 10,
            'private_repos': 20
        }
    })

    def test_random_username(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_CLEAN_USERNAMES': False,
            'SOCIAL_AUTH_SLUGIFY_USERNAMES': True
        })
        self.do_login()


class RepeatedUsernameTest(BaseActionTest):
    def test_random_username(self):
        User(username='foobar')
        self.do_login(after_complete_checks=False)
        self.assertTrue(self.strategy.session_get('username')
                                     .startswith('foobar'))


class AssociateByEmailTest(BaseActionTest):
    def test_multiple_accounts_with_same_email(self):
        user = User(username='foobar1')
        user.email = 'foo@bar.com'
        self.do_login(after_complete_checks=False)
        self.assertTrue(self.strategy.session_get('username')
                                     .startswith('foobar'))


class MultipleAccountsWithSameEmailTest(BaseActionTest):
    def test_multiple_accounts_with_same_email(self):
        user1 = User(username='foobar1')
        user2 = User(username='foobar2')
        user1.email = 'foo@bar.com'
        user2.email = 'foo@bar.com'
        with self.assertRaises(AuthException):
            self.do_login(after_complete_checks=False)


class UserPersistsInPartialPipeline(BaseActionTest):
    def test_user_persists_in_partial_pipeline_kwargs(self):
        user = User(username='foobar1')
        user.email = 'foo@bar.com'

        self.strategy.set_settings({
            'SOCIAL_AUTH_PIPELINE': (
                'social_core.pipeline.social_auth.social_details',
                'social_core.pipeline.social_auth.social_uid',
                'social_core.pipeline.social_auth.associate_by_email',
                'social_core.tests.pipeline.set_user_from_kwargs'
            )
        })

        self.do_login(after_complete_checks=False)

        # Handle the partial pipeline
        self.strategy.session_set('attribute', 'testing')
        token = self.strategy.session_pop(PARTIAL_TOKEN_SESSION_NAME)
        partial = self.strategy.partial_load(token)
        self.backend.continue_pipeline(partial)

    def test_user_persists_in_partial_pipeline(self):
        user = User(username='foobar1')
        user.email = 'foo@bar.com'

        self.strategy.set_settings({
            'SOCIAL_AUTH_PIPELINE': (
                'social_core.pipeline.social_auth.social_details',
                'social_core.pipeline.social_auth.social_uid',
                'social_core.pipeline.social_auth.associate_by_email',
                'social_core.tests.pipeline.set_user_from_args'
            )
        })

        self.do_login(after_complete_checks=False)

        # Handle the partial pipeline
        self.strategy.session_set('attribute', 'testing')
        token = self.strategy.session_pop(PARTIAL_TOKEN_SESSION_NAME)
        partial = self.strategy.partial_load(token)
        self.backend.continue_pipeline(partial)



"""
Open Suse OpenId backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/suse.html
"""
from .open_id import OpenIdAuth


class OpenSUSEOpenId(OpenIdAuth):
    name = 'opensuse'
    URL = 'https://www.opensuse.org/openid/user/'

    def get_user_id(self, details, response):
        """
        Return user unique id provided by service. For openSUSE
        the nickname is original.
        """
        return details['nickname']



"""
Tumblr OAuth1 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/tumblr.html
"""
from ..utils import first
from .oauth import BaseOAuth1


class TumblrOAuth(BaseOAuth1):
    name = 'tumblr'
    ID_KEY = 'name'
    AUTHORIZATION_URL = 'http://www.tumblr.com/oauth/authorize'
    REQUEST_TOKEN_URL = 'http://www.tumblr.com/oauth/request_token'
    REQUEST_TOKEN_METHOD = 'POST'
    ACCESS_TOKEN_URL = 'http://www.tumblr.com/oauth/access_token'

    def get_user_id(self, details, response):
        return response['response']['user'][self.ID_KEY]

    def get_user_details(self, response):
        # http://www.tumblr.com/docs/en/api/v2#user-methods
        user_info = response['response']['user']
        data = {'username': user_info['name']}
        blog = first(lambda blog: blog['primary'], user_info['blogs'])
        if blog:
            data['fullname'] = blog['title']
        return data

    def user_data(self, access_token):
        return self.get_json('http://api.tumblr.com/v2/user/info',
                             auth=self.oauth_auth(access_token))



"""
Coinbase OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/coinbase.html
"""
from .oauth import BaseOAuth2


class CoinbaseOAuth2(BaseOAuth2):
    name = 'coinbase'
    SCOPE_SEPARATOR = '+'
    DEFAULT_SCOPE = ['user', 'balance']
    AUTHORIZATION_URL = 'https://www.coinbase.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://api.coinbase.com/oauth/token'
    REVOKE_TOKEN_URL = 'https://api.coinbase.com/oauth/revoke'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False

    def get_user_id(self, details, response):
        return response['data']['id']

    def get_user_details(self, response):
        """Return user details from Coinbase account"""
        user_data = response['data']
        email = user_data.get('email', '')
        name = user_data['name']
        username = user_data.get('username')
        fullname, first_name, last_name = self.get_user_names(name)
        return {'username': username,
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name,
                'email': email}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json('https://api.coinbase.com/v2/user',
                headers={'Authorization': 'Bearer ' + access_token})



"""
Chatwork OAuth2 backend
"""
import base64

from .oauth import BaseOAuth2


class ChatworkOAuth2(BaseOAuth2):
    """Chatwork OAuth authentication backend"""
    name = 'chatwork'
    API_URL = 'https://api.chatwork.com/v2'
    AUTHORIZATION_URL = 'https://www.chatwork.com/packages/oauth2/login.php'
    ACCESS_TOKEN_URL = 'https://oauth.chatwork.com/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = True
    DEFAULT_SCOPE = ['users.profile.me:read']
    ID_KEY = 'account_id'
    EXTRA_DATA = [
        ('expires_in', 'expires'),
        ('refresh_token', 'refresh_token')
    ]

    def api_url(self, path):
        api_url = self.setting('API_URL') or self.API_URL
        return '{0}{1}'.format(api_url.rstrip('/'), path)

    def auth_headers(self):
        return {
            'Authorization': b'Basic ' + base64.b64encode(
                '{0}:{1}'.format(*self.get_key_and_secret()).encode()
            )
        }

    def auth_complete_params(self, state=None):
        return {
            'grant_type': 'authorization_code',
            'code': self.data.get('code', ''),
            'redirect_uri': self.get_redirect_uri(state)
        }

    def get_user_details(self, response):
        """Return user details from Chatwork account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get('name')
        )
        username = response.get('chatwork_id') or \
                   response.get('login_mail') or \
                   response.get('account_id')
        email = response.get('mail') or \
                response.get('login_mail') or \
                ''
        return {
            'username': username,
            'email': email,
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        headers = {'Authorization': 'Bearer ' + access_token}
        return self.get_json(self.api_url('/me'), headers=headers)

    def refresh_token_params(self, token, *args, **kwargs):
        return {'refresh_token': token, 'grant_type': 'refresh_token'}



"""
Yandex OpenID and OAuth2 support.

This contribution adds support for Yandex.ru OpenID service in the form
openid.yandex.ru/user. Username is retrieved from the identity url.

If username is not specified, OpenID 2.0 url used for authentication.
"""
from six.moves.urllib_parse import urlsplit

from .open_id import OpenIdAuth
from .oauth import BaseOAuth2


class YandexOpenId(OpenIdAuth):
    """Yandex OpenID authentication backend"""
    name = 'yandex-openid'
    URL = 'http://openid.yandex.ru'

    def get_user_id(self, details, response):
        return details['email'] or response.identity_url

    def get_user_details(self, response):
        """Generate username from identity url"""
        values = super(YandexOpenId, self).get_user_details(response)
        values['username'] = values.get('username') or\
                             urlsplit(response.identity_url)\
                                    .path.strip('/')
        values['email'] = values.get('email', '')
        return values


class YandexOAuth2(BaseOAuth2):
    """Legacy Yandex OAuth2 authentication backend"""
    name = 'yandex-oauth2'
    AUTHORIZATION_URL = 'https://oauth.yandex.com/authorize'
    ACCESS_TOKEN_URL = 'https://oauth.yandex.com/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(
            response.get('real_name') or response.get('display_name') or ''
        )
        email = response.get('default_email')
        if not email:
            emails = response.get('emails')
            email = emails[0] if emails else ''
        return {'username': response.get('display_name'),
                'email': email,
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json('https://login.yandex.ru/info',
                             params={'oauth_token': access_token,
                                     'format': 'json'})


class YaruOAuth2(BaseOAuth2):
    name = 'yaru'
    AUTHORIZATION_URL = 'https://oauth.yandex.com/authorize'
    ACCESS_TOKEN_URL = 'https://oauth.yandex.com/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(
            response.get('real_name') or response.get('display_name') or ''
        )
        email = response.get('default_email')
        if not email:
            emails = response.get('emails')
            email = emails[0] if emails else ''
        return {'username': response.get('display_name'),
                'email': email,
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json('https://login.yandex.ru/info',
                             params={'oauth_token': access_token,
                                     'format': 'json'})



"""
Zotero OAuth1 backends, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/zotero.html
"""
from .oauth import BaseOAuth1


class ZoteroOAuth(BaseOAuth1):

    """Zotero OAuth authorization mechanism"""
    name = 'zotero'
    AUTHORIZATION_URL = 'https://www.zotero.org/oauth/authorize'
    REQUEST_TOKEN_URL = 'https://www.zotero.org/oauth/request'
    ACCESS_TOKEN_URL = 'https://www.zotero.org/oauth/access'

    def get_user_id(self, details, response):
        """
        Return user unique id provided by service. For Ubuntu One
        the nickname should be original.
        """
        return details['userID']

    def get_user_details(self, response):
        """Return user details from Zotero API account"""
        access_token = response.get('access_token', {})
        return {
            'username': access_token.get('username', ''),
            'userID': access_token.get('userID', '')
        }



"""
LoginRadius BaseOAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/loginradius.html
"""
from .oauth import BaseOAuth2


class LoginRadiusAuth(BaseOAuth2):
    """LoginRadius BaseOAuth2 authentication backend."""
    name = 'loginradius'
    ID_KEY = 'ID'
    ACCESS_TOKEN_URL = 'https://api.loginradius.com/api/v2/access_token'
    PROFILE_URL = 'https://api.loginradius.com/api/v2/userprofile'
    ACCESS_TOKEN_METHOD = 'GET'
    REDIRECT_STATE = False
    STATE_PARAMETER = False

    def uses_redirect(self):
        """Return False because we return HTML instead."""
        return False

    def auth_html(self):
        key, secret = self.get_key_and_secret()
        tpl = self.setting('TEMPLATE', 'loginradius.html')
        return self.strategy.render_html(tpl=tpl, context={
            'backend': self,
            'LOGINRADIUS_KEY': key,
            'LOGINRADIUS_REDIRECT_URL': self.get_redirect_uri()
        })

    def request_access_token(self, *args, **kwargs):
        return self.get_json(params={
            'token': self.data.get('token'),
            'secret': self.setting('SECRET')
        }, *args, **kwargs)

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service. Implement in subclass."""
        return self.get_json(
            self.PROFILE_URL,
            params={'access_token': access_token},
            data=self.auth_complete_params(self.validate_state()),
            headers=self.auth_headers(),
            method=self.ACCESS_TOKEN_METHOD
        )

    def get_user_details(self, response):
        """Must return user details in a know internal struct:
            {'username': <username if any>,
             'email': <user email if any>,
             'fullname': <user full name if any>,
             'first_name': <user first name if any>,
             'last_name': <user last name if any>}
        """
        profile = {
            'username': response['NickName'] or '',
            'email': response['Email'][0]['Value'] or '',
            'fullname': response['FullName'] or '',
            'first_name': response['FirstName'] or '',
            'last_name': response['LastName'] or ''
        }
        return profile

    def get_user_id(self, details, response):
        """Return a unique ID for the current user, by default from server
        response. Since LoginRadius handles multiple providers, we need to
        distinguish them to prevent conflicts."""
        return '{0}-{1}'.format(response.get('Provider'),
                                response.get(self.ID_KEY))



"""
Sketchfab OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/sketchfab.html
    https://sketchfab.com/developers/oauth
"""
from .oauth import BaseOAuth2


class SketchfabOAuth2(BaseOAuth2):
    name = 'sketchfab'
    ID_KEY = 'uid'
    AUTHORIZATION_URL = 'https://sketchfab.com/oauth2/authorize/'
    ACCESS_TOKEN_URL = 'https://sketchfab.com/oauth2/token/'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False
    REQUIRES_EMAIL_VALIDATION = False
    EXTRA_DATA = [
        ('username', 'username'),
        ('apiToken', 'apiToken')
    ]

    def get_user_details(self, response):
        """Return user details from Sketchfab account"""
        user_data = response
        email = user_data.get('email', '')
        username = user_data['username']
        name = user_data.get('displayName', '')
        fullname, first_name, last_name = self.get_user_names(name)
        return {'username': username,
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name,
                'email': email}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json('https://sketchfab.com/v2/users/me', headers={
            'Authorization': 'Bearer {0}'.format(access_token)
        })



"""
Clef OAuth support.

This contribution adds support for Clef OAuth service. The settings
SOCIAL_AUTH_CLEF_KEY and SOCIAL_AUTH_CLEF_SECRET must be defined with the
values given by Clef application registration process.
"""

from .oauth import BaseOAuth2


class ClefOAuth2(BaseOAuth2):
    """Clef OAuth authentication backend"""
    name = 'clef'
    AUTHORIZATION_URL = 'https://clef.io/iframes/qr'
    ACCESS_TOKEN_URL = 'https://clef.io/api/v1/authorize'
    ACCESS_TOKEN_METHOD = 'POST'
    SCOPE_SEPARATOR = ','

    def auth_params(self, *args, **kwargs):
        params = super(ClefOAuth2, self).auth_params(*args, **kwargs)
        params['app_id'] = params.pop('client_id')
        params['redirect_url'] = params.pop('redirect_uri')
        return params

    def get_user_id(self, response, details):
        return details.get('info').get('id')

    def get_user_details(self, response):
        """Return user details from Github account"""
        info = response.get('info')
        fullname, first_name, last_name = self.get_user_names(
            first_name=info.get('first_name'),
            last_name=info.get('last_name')
        )

        email = info.get('email', '')
        if email:
            username = email.split('@', 1)[0]
        else:
            username = info.get('id')

        return {
            'username': username,
            'email': email,
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name,
            'phone_number': info.get('phone_number', '')
        }

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json('https://clef.io/api/v1/info',
                             params={'access_token': access_token})



from social_core.backends.oauth import BaseOAuth2


class AtlassianOAuth2(BaseOAuth2):
    name = 'atlassian'
    AUTHORIZATION_URL = 'https://accounts.atlassian.com/authorize'
    ACCESS_TOKEN_METHOD = 'POST'
    ACCESS_TOKEN_URL = 'https://api.atlassian.com/oauth/token'
    DEFAULT_SCOPE = ['read:jira-user', 'offline_access']
    ID_KEY = 'accountId'
    EXTRA_DATA = [
        ('resources', 'resources'),
        ('refresh_token', 'refresh_token'),
        ('expires_in', 'expires_in'),
    ]

    def auth_params(self, state=None):
        params = super(AtlassianOAuth2, self).auth_params(state)
        params.update({'audience': 'api.atlassian.com',
                       'prompt': 'consent'})
        return params

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(response['displayName'])
        return {'username': response['name'],
                'email': response['emailAddress'],
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        resources = self.get_json('https://api.atlassian.com/oauth/token/accessible-resources',
                                  headers={'Authorization': 'Bearer {}'.format(access_token)})
        user_info = self.get_json('https://api.atlassian.com/ex/jira/{}/rest/api/2/myself'.format(resources[0]['id']),
                                  headers={'Authorization': 'Bearer {}'.format(access_token)})
        user_info['resources'] = resources
        return user_info
